#' @title Plot locations of basis function centers
#'
#' @description This
#' plot may be considered an incomplete map--- it includes just the locations of
#' basis function centers and the domain boundary. It can made complete by
#' adding, for example, geographic information and a suitable projection;
#' this information is added using the functions in library MapUtilities.
#'
#' @param x
#' Object of class BasisFunctions.
#'
#' @param spacing
#' Character string containing the spacing between the basis function
#' centers. This spacing must have been specfied when the class BasisFunctions
#' was instantiated.
#'
#' @param ...
#' Required for compatibility with generic definition
#'
#' @return A ggplot2 object to implement the previously-described plotting.
#'
#' @export
#'
plot.BasisFunctions <- function(x, spacing, ...) {

  spacing <- as.integer(spacing)

  index <- which(x$file_info$center_spacings == spacing)
  if(length(index) == 0){
    stop( sprintf("The spacing of %g km was not used to compute any basis functions.\n",
                  spacing),
          call. = FALSE )
  }

  load(x$file_info$filenames[index])

  ggplot2::ggplot() +
    MappingUtilities::Add_Points(as.data.frame(basis_functions$centers_longlat), shape = 16) +
    MappingUtilities::Add_Path(x$domain_boundary, "blue") +
    ggplot2::ggtitle(paste0("Spacing: ", spacing, " kilometers"))
}

#' @title Summarize the basis functions
#'
#' @description Construct a table that summarizes, at each of the spacings,
#' pertinent details of the matrix that stores the values of the basis
#' functions at the sample locations.
#'
#' @param object
#' Object of class BasisFunctions.
#'
#' @param ...
#' Required for compatibility with generic definition
#'
#' @details
#' The table includes, for each spacing, the number of rows, the number of
#' columns, the number of non-zero elements, and the percentage of all matrix
#' elements that are non-zero.
#'
#' @export
#'
summary.BasisFunctions <- function(object, ...) {

  cat( sprintf( "####################################################################\n"))
  cat(sprintf("Summary\n"))
  cat(sprintf("                                       Number of    Percentage of\n"))
  cat(sprintf("             Number of    Number of    non-zero     all elements that\n"))
  cat(sprintf("Spacing      rows         columns      elements     are non-zero\n"))

  for(i in 1:nrow(object$file_info)){
    load(object$file_info$filenames[i])
    cat(sprintf("%-10.3g   %-10d   %-10d   %-10d   %-10.3g\n",
                object$file_info$center_spacings[i],
                nrow(basis_functions$basisFuncs),
                ncol(basis_functions$basisFuncs),
                length(basis_functions$basisFuncs@x),
                100 * length(basis_functions$basisFuncs@x) /
                  (nrow(basis_functions$basisFuncs) * ncol(basis_functions$basisFuncs))))
  }

  cat( sprintf( "####################################################################\n"))
}


# @title BasisFuncMatrix
#
# @description Calculate the matrix that contains the values of the basis
# functions at the sample locations
#
# @param centers
# Object of class sp::SpatialPoints containing the locations
#        of the centers of the basis functions. Locations must be specifed in
#        a Cartesian coordinate system.
#
# @param sampleLocs
# Object of class sp::SpatialPoints containing the
#        locations of the field samples. Locations must be specifed in
#        a Cartesian coordinate system.
#
# @param sepDistance
# Distance between the centers of the basis functions.
#
# @param estNNZ
# Estimated maximum number of non-zero elements in the matrix that
#        contains the values of the basis functions.
#
# @details
# The units for \code{centers}, \code{sampleLocs}, and \code{sepDistance} must
# be the same.
#
# This function generates basis functions using the local
# bisquare function (Cressie and Johannesson, 2008).
#
# @return The matrix that contains  the values of the basis
# functions at the sample locations. Each column of the matrix pertains to
# one basis function. (Thus, the
# number of columns in this matrix equals the number of rows in matrix
# centers_longlat.) Each row of the matrix comprises the values of the basis
# functions at the sample locations. (Thus, the number of rows in this matrix
# equals the number of rows in argument sample_locations.) The
# matrix is sparse and is stored in the "dictionary-of-keys" format, which is
# implemented with class Matrix::dgTMatrix.
#
# @references
# Cressie, N., and Johnnesson, G., 2008, Fixed rank kriging for very large
# spatial data sets: Journal of the Royal Statistical Society, Section B,
# vol. 70, Part 1, p. 209-226.
#
BasisFuncMatrix <- function(centers, sampleLocs, sepDistance,
                           estNNZ) {

  CalcDist <- function( coords1, coords2 ) {
    dx <- coords1[1] - coords2[, 1]
    dy <- coords1[2] - coords2[, 2]
    return( sqrt( dx^2 + dy^2 ) )
  }

  centers_coord <- sp::coordinates(centers)
  data_coord <- sp::coordinates(sampleLocs)

  r_max <- sepDistance * 1.5

  J <- nrow(centers_coord)
  I <- nrow(data_coord)

  row_indices <- integer(estNNZ)
  col_indices <- integer(estNNZ)
  elements <- numeric(estNNZ)

  NNZ <- 0  # number of non-zero elements
  for( j in 1:J) {
    distances <- CalcDist(centers_coord[j, ], data_coord)

    areNonZero <- distances <= r_max
    M <- sum(areNonZero)
    if(M > 0){
      if(NNZ > estNNZ){
        stop( sprintf("The number of non-zero elements in the matrix for the\n"),
              sprintf("basis functions is greater than %d.\n", estNNZ),
              sprintf("Increase argument estNNZ.\n"),
              call. = FALSE )
      }
      row_indices[(NNZ+1):(NNZ+M)] <- which(areNonZero)
      col_indices[(NNZ+1):(NNZ+M)] <- rep.int(j, M)
      elements[(NNZ+1):(NNZ+M)] <- (1 - (distances[areNonZero] / r_max)^2)^2
      NNZ <- NNZ + M
    }
    # else {
    #   stop( sprintf("For basis function %d, for which the center is at (%g, %g),\n",
    #                 j, centers_coord[j, 1], centers_coord[j, 2]),
    #         sprintf("all function values are zero.\n"),
    #         call. = FALSE )
    # }

  }

  # Store the matrix in the "dictionary-of-keys" format.
  # S <- new("dgTMatrix",
  #           i = row_indices[1:NNZ] - 1L,
  #           j = col_indices[1:NNZ] - 1L,
  #           x = elements[1:NNZ],
  #           Dim = c(I, J))

  # Store the matrix in the "dictionary-of-keys" format.
  S <- Matrix::spMatrix(I, J,
                        i = row_indices[1:NNZ],
                        j = col_indices[1:NNZ],
                        x = elements[1:NNZ])
  return(S)
}

# @title Prune
#
# @description Prune the matrix of basis functions and the associated centers.
#
# @param basisFuncs
# Matrix containing the values of the basis functions at the sample
# locations. It is the value of function BasisFuncMatrix.
#
# @param centers
# Object of class sp::SpatialPoints containing the locations
# of the centers of the basis functions. Locations must be specifed in
# a Cartesian coordinate system.
#
# @param domain
# Object of class sp::SpatialPolygons containing the vertices
# of the polygon that defines the boundary of the domain.
# The locations of the vertices must be specifed in
# a Cartesian coordinate system.
#
# @param minFuncValue
# See details.
#
# @param minNumSamples
# See details.
#
# @param areOutsideRemoved
# See details.
#
# @details
# The units for \code{centers} and \code{domain} must be the same.
#
# In the Bayesian modeling, the weight applied to the basis function must be
# estimated. For this weight to be identifable, there must be enough samples
# with large enough function values. The lower threshold for the minimum
# number of samples is specified by argument minNumSamples; the default value
# of 5 seems adequate. The lower threshold for the function value is specified
# by argument  minFuncValue; the default value of 0.1 seems adequate.
# If it is necessary to remove centers that are outside the domain, set argument
# areOutsideRemoved to TRUE. Basis functions and the associated centers that don't meet this criteria are
# are removed (pruned).
#
# @return A list with the following components is returned.
# @return \item{basisFuncs}{Matrix of basis functions after pruning. The
# matrix is stored in the "dictionary-of-keys" format, which is
# implemented with class Matrix::dgTMatrix.}
# @return \item{centers}{Object of class sp::SpatialPoints containing the
# locations of the centers; these centers are associated with the basis
# functions in \code{basisFuncs}.}
#
Prune <- function(basisFuncs, centers, domain,
                  minFuncValue,
                  minNumSamples,
                  areOutsideRemoved){

  J <- ncol(basisFuncs)
  areOK <- rep.int(TRUE, J)
  for(j in 1:J){
    if(sum(basisFuncs[, j] > minFuncValue) < minNumSamples) {
      areOK[j] <- FALSE
    }
  }

  if(areOutsideRemoved){
    point <- sp::coordinates(centers)
    pol <- domain@polygons[[1]]@Polygons[[1]]@coords
    result <- sp::point.in.polygon(point[, 1], point[, 2], pol[, 1], pol[, 2])
    areOutside <- result == 0L
    areOK <- areOK & !areOutside
  }

  return(list(basisFuncs = basisFuncs[, areOK],
              centers = centers[areOK, ]))
}


# @title CalcBasisFunctions
#
# @description Calculate the basis functions for one spacing between the
# function centers
#
# @param sepDistance
# Distance (in meters) between the centers of the basis functions.
#
# @param domain_boundary
# Dataframe containing the longitudes and latitudes of the
#              points that define the boundary of the domain. The first column
#              of the dataframe lists the longitudes, and the second column
#              the latitudes. The column names are "long" and "lat".
#
# @param sample_locations
# Dataframe containing the longitudes and latitudes of the sample
#              locations. The column name for the longitudes is "long", and
#              the column name for the latitudes is "lat".
#
# @param CRS_arg_longlat
# Character string containg the arguments to function sp::CRS that
# specfy the projection for arguments domain_boundary and sample_locations.
#
# @param CRS_arg_utm
# Character string containg the arguments to function sp::CRS that
# specify the projection, in universal transverse mercator (utm) coordinates,
# for arguments domain_boundary and sample_locations.
#
# @param seed
# Random number seed.
#
# @param prune_minFuncValue
# See details.
#
# @param prune_minNumberSamples
# See details.
#
# @param prune_areOutsideRemoved
# See details.
#
# @param estNNZ
# Estimated maximum number of non-zero elements in the matrix that contains the
# basis functions.
#
# @details
# Calculating the basis functions involves calculating both the locations
# of their centers and the values of the basis functions at the sample locations.
#
# The seed may be set to a natural number to make the calculations reproducible.
#
# In the Bayesian modeling, the weight applied to the basis function must be
# estimated. For this weight to be identifable, there must be enough samples
# with large enough function values. The lower threshold for the minimum
# number of samples is specified by argument prune_minNumSamples; the default value
# of 5 seems adequate. The lower threshold for the function value is specified
# by argument  prune_minFuncValue; the default value of 0.1 seems adequate.
# If it is necessary to remove centers that are outside the domain, set argument
# prune_areOutsideRemoved to TRUE.
#
# @return A list with the following components is returned.
# @return \item{sepDistance}{Argument sepDistance.}
# @return \item{CRS_arg_longlat}{Argument CRS_arg_longlat.}
# @return \item{CRS_arg_utm}{Argument CRS_arg_utm.}
# @return \item{seed}{Argument seed.}
# @return \item{prune_minFuncValue}{Argument prune_minFuncValue.}
# @return \item{prune_minNumSamples}{Argument prune_minNumSamples.}
# @return \item{prune_areOutsideRemoved}{Argument prune_areOutsideRemoved.}
# @return \item{centers_longlat}{Matrix containing the longitudes and latitudes
# of the (basis function) centers. Each row in the
# matrix pertains the center of one basis function. The first column lists
# the longitudes, and the second lists the latitudes. The datum for these
# coordinates is specified in argument CRS_arg_longlat.}
# @return \item{centers_utm}{Matrix containing the utm coordinates
# of the (basis function) centers. Each row in the
# matrix pertains the center of one basis function. The first column lists
# the easting in meters, and the second lists the northing in meters.
# The datum for these coordinates is specified in argument CRS_arg_utm.}
# @return \item{basisFuncs}{Matrix containing the values of the basis
# functions at the sample locations. Each column of the matrix pertains to
# one basis function. (Thus, the
# number of columns in this matrix equals the number of rows in matrix
# centers_longlat.) Each row of the matrix comprises the values of the basis
# functions at the sample locations. (Thus, the number of rows in this matrix
# equals the number of rows in argument sample_locations.) The
# matrix is sparse and is stored in the "dictionary-of-keys" format, which is
# implemented with class Matrix::dgTMatrix.}
#
CalcBasisFunctions <- function(sepDistance,
                           domain_boundary,
                           sample_locations,
                           CRS_arg_longlat,
                           CRS_arg_utm,
                           seed = NULL,
                           prune_minFuncValue = 0.1,
                           prune_minNumSamples = 5,
                           prune_areOutsideRemoved = FALSE,
                           estNNZ = as.integer(1e7)) {

  # recast the domain boundary as a spatial polygon container
  tmp <- as.matrix(domain_boundary)
  sp_domain_boundary <- sp::SpatialPolygons(list(sp::Polygons(list(sp::Polygon(tmp)), "1")),
                                  proj4string = sp::CRS(CRS_arg_longlat))


  # transform the lats/longs to utm
  sp_domain_boundary_utm <- sp::spTransform(sp_domain_boundary, sp::CRS(CRS_arg_utm) )

  # specify a bounding box for the domain boundary
  a <- sp::bbox(sp_domain_boundary_utm)
  margin <- 2 * sepDistance
  coords <- rbind(
    c(a[1,1]-margin, a[2,1]-margin),
    c(a[1,1]-margin, a[2,2]+margin),
    c(a[1,2]+margin, a[2,2]+margin),
    c(a[1,2]+margin, a[2,1]-margin)
  )

  box_utm <- sp::SpatialPolygons(list(sp::Polygons(list(sp::Polygon(coords)), "1")),
                             proj4string = sp::CRS(CRS_arg_utm))

  # calculate the centers of the basis functions within the box
  set.seed(seed)
  centers_utm <- sp::spsample(box_utm, type = "hexagonal",
                              cellsize = sepDistance)

  # calculate the basis functions at the sample locations
  sp_sample_locations <- sp::SpatialPoints(sample_locations[, c("long","lat"), drop = FALSE],
                                proj4string=sp::CRS(CRS_arg_longlat))

  sp_sample_locations_utm <- sp::spTransform( sp_sample_locations,
                                              sp::CRS(CRS_arg_utm) )

  basisFuncs <- BasisFuncMatrix(centers_utm, sp_sample_locations_utm,
                                sepDistance, estNNZ)

  # Prune the centers and the basis functions
  tmp <- Prune(basisFuncs, centers_utm, sp_domain_boundary_utm,
               prune_minFuncValue,
               prune_minNumSamples,
               prune_areOutsideRemoved)
  centers_utm <- tmp$centers
  basisFuncs <- tmp$basisFuncs

  # transform the centers to long/lat
  centers_longlat <- sp::spTransform(centers_utm, sp::CRS(CRS_arg_longlat))

  # only the coordinates of the centers are needed
  centers_utm <- sp::coordinates(centers_utm)
  colnames(centers_utm) <- c("Easting", "Northing")

  centers_longlat <- sp::coordinates(centers_longlat)
  colnames(centers_longlat) <- c("long", "lat")

  return(list(sepDistance = sepDistance,
              CRS_arg_longlat = CRS_arg_longlat,
              CRS_arg_utm = CRS_arg_utm,
              seed = seed,
              prune_minFuncValue = prune_minFuncValue,
              prune_minNumSamples = prune_minNumSamples,
              prune_areOutsideRemoved = prune_areOutsideRemoved,
              basisFuncs = basisFuncs,
              centers_utm = centers_utm,
              centers_longlat = centers_longlat))
}

#' @title Constructor of class to generate basis functions.
#'
#' @description This constructor generates a set of basis functions for each
#' specfied spacing between the basis function centers. Each set is written
#' to a file in a common directory.
#'
#' @param spacings
#' Vector of character strings. Each element contains the spacing
#' (in kilometers) between the basis function centers.
#'
#' @param domain_boundary
#' Dataframe containing the longitudes and latitudes of the
#'              points that define the boundary of the domain. The first column
#'              of the dataframe lists the longitudes, and the second column
#'              the latitudes. The column names are "long" and "lat".
#'
#' @param spatialData
#' Container with the spatial data and associated information.
#' See the User's Guide for a detailed description of this container.
#'
#' @param CRS_arg_utm
#' Character string containing the arguments to function sp::CRS that
#' specify the projection, in universal transverse mercator (utm) coordinates,
#' for arguments domain_boundary and spatialData.
#'
#' @param directory_name
#' Character string containing the name of the directory that will store
#' the files with the basis functions.
#'
#' @param seed
#' Random number seed.
#'
#' @param prune_minFuncValue
#' See details.
#'
#' @param prune_minNumSamples
#' See details.
#'
#' @param prune_areOutsideRemoved
#' See details.
#'
#' @param estNNZ
#' Estimated maximum number of non-zero elements in the matrix that contains the
#' basis functions.
#'
#' @details
#' In the Bayesian modeling, the weight applied to the basis function must be
#' estimated. For this weight to be identifable, there must be enough samples
#' with large enough function values. The lower threshold for the minimum
#' number of samples is specified by \code{prune_minNumSamples}; the default value
#' of 5 seems adequate. The lower threshold for the function value is specified
#' by \code{prune_minFuncValue}; the default value of 0.1 seems adequate.
#' If it is necessary to remove centers that are outside the domain, set argument
#' prune_areOutsideRemoved to TRUE.
#'
#' For each element of \code{spacings}, an R-language list is constructed that
#' includes the spacing,
#' argument \code{CRS_arg_longlat},
#' argument \code{CRS_arg_utm},
#' argument \code{seed},
#' argument \code{prune_minFuncValue},
#' argument \code{prune_minNumSamples},
#' argument \code{prune_areOutsideRemoved},
#' matrix \code{centers_utm},
#' matrix \code{centers_longlat},
#' and matrix \code{BasisFuncs}.
#'
#' \itemize{
#' \item Matrix \code{centers_utm} contains the utm coordinates
#' of the (basis function) centers. Each row in the
#' matrix pertains the center of one basis function. The first column lists
#' the easting in meters, and the second lists the northing in meters.
#' The datum for these coordinates is specified in argument \code{CRS_arg_utm}.
#' The calculation of these centers is partly random; to make it
#' reproducible, argument \code{seed} must be set to a natural number.
#'
#' \item Matrix \code{centers_longlat} contains the longitudes and latitudes
#' of the (basis function) centers. Each row in the
#' matrix pertains the center of one basis function. The first column lists
#' the longitudes, and the second lists the latitudes. The datum for these
#' coordinates is specified in argument \code{CRS_arg_longlat}. These latitudes
#' and longitudes are calculated from the utm coordinates in matrix
#' \code{centers_utm}.
#'
#' \item Matrix \code{BasisFuncs} contains the values of the basis
#' functions at the sample locations, which are specified in
#' argument \code{spatialData}.
#' The basis function itself is the local
#' bisquare function (Cressie and Johannesson, 2008).
#' Each column of the matrix pertains to
#' one basis function. (Thus, the
#' number of columns in this matrix equals the number of rows in matrix
#' centers_longlat.) Each row of the matrix comprises the values of the basis
#' functions at the sample locations. (Thus, the number of rows in this matrix
#' equals the number of sample locations in \code{spatialData}.) The
#' matrix is sparse and is stored in the "dictionary-of-keys" format, which is
#' implemented with class Matrix::dgTMatrix.
#' }
#'
#' This list is
#' written to a file, for which the name includes the spacing between the
#' basis function centers. This file is placed in
#' directory \code{directory_name}.
#'
#' @return A list with the following components is returned.
#' @return \item{spacings}{Argument spacings.}
#' @return \item{domain_boundary}{Argument domain_boundary.}
#' @return \item{estNNZ}{Argument estNNZ.}
#' @return \item{file_info}{R-language tibble containing the names of the
#' files storing the basis functions and the associated spacings between the
#' function centers.}
#' @return \item{call}{Function call.}
#'
#' @references
#' Cressie, N., and Johnnesson, G., 2008, Fixed rank kriging for very large
#' spatial data sets: Journal of the Royal Statistical Society, Section B,
#' vol. 70, Part 1, p. 209-226.
#'
#' @export
#'
BasisFunctions <- function(spacings,
                           domain_boundary,
                           spatialData,
                           CRS_arg_utm,
                           directory_name = "BasisFunctions",
                           seed = NULL,
                           prune_minFuncValue = 0.1,
                           prune_minNumSamples = 5,
                           prune_areOutsideRemoved = FALSE,
                           estNNZ = as.integer(1e7)) {

  file_info <- tibble::tibble(filenames = paste0(directory_name, "\\Spacing_", spacings, ".dat"),
                                 center_spacings = as.integer(spacings))

  sample_locations <- as.data.frame(sp::coordinates(spatialData))

  CRS_arg_longlat <- sp::proj4string(spatialData)

  if(!dir.exists(directory_name)){
    dir.create(directory_name)
  }

  for(i in 1:nrow(file_info)){

    basis_functions <- CalcBasisFunctions(1000 * file_info$center_spacings[i],
                                      domain_boundary,
                                      sample_locations,
                                      CRS_arg_longlat,
                                      CRS_arg_utm,
                                      seed = seed,
                                      prune_minFuncValue = prune_minFuncValue,
                                      prune_minNumSamples = prune_minNumSamples,
                                      estNNZ = estNNZ)

    save(basis_functions, file = file_info$filenames[i])
    cat(sprintf("Finished spacing %d\n", file_info$center_spacings[i]))
  }

  rval <- list(spacings = spacings,
               domain_boundary = domain_boundary,
               estNNZ = estNNZ,
               file_info = file_info,
               call = sys.call())

  class(rval) <- "BasisFunctions"

  return(rval)
}


