#' @title Graphical check of the standardized residuals
#'
#' @description
#' The graphical check involves three plots:
#' (1) A histogram showing the distribution of the standarized residuals.
#' (2) A quantile-quantile plot comparing the standarized residuals to
#' quantiles from a standardized normal distribution.
#' (3) A semivariogram showing the variance between pairs of standardized
#' residuals as a function of the distance between them.
#'
#' @param x
#' Object of class CheckResiduals.
#'
#' @param spatialData
#' Container with the spatial data and associated information.
#' See the User's Guide for a detailed description of this container.
#'
#' @param CRS_arg_utm
#' Character string containing the arguments to function sp::CRS that
#' specify the projection, in universal transverse mercator (utm) coordinates,
#' for argument \code{spatialData}.
#'
#' @param variogram_breaks
#' Real-valued vector specifying the distances (in kilometers) for the
#' variogram bins.
#'
#' @param ...
#' Required for compatibility with generic definition
#'
#' @export
#'
plot.CheckResiduals <- function(x, spatialData, CRS_arg_utm,
                                variogram_breaks = NULL, ...){

  p1 <- ggplot2::ggplot(data = x$residuals) +
    ggplot2::geom_histogram(ggplot2::aes(x = std_residual, y = ..density..),
                            fill = "gray50") +
    ggplot2::scale_x_continuous(name = "Standardized residual") +
    ggplot2::scale_y_continuous(name = "Density") +
    ggplot2::ggtitle("A.") +
    ggplot2::theme_bw() +
    ggplot2::theme(panel.grid.major = ggplot2::element_blank(),
                   panel.grid.minor = ggplot2::element_blank()) +
    ggplot2::theme(plot.title=ggplot2::element_text(hjust = 0,face = "bold", size = 10))


  df1 <- tibble::tibble(std_residual = sort(x$residuals$std_residual),
                       prob = stats::ppoints(length(std_residual)),
                       theoretical_quantile = stats::qnorm(prob))

  theRange = range(df1[, -2])

  p2 <- ggplot2::ggplot(data = df1) +
    ggplot2::geom_abline(slope = 1, color = "green") +
    ggplot2::geom_point(ggplot2::aes(x = theoretical_quantile, y = std_residual),
               shape = 16, color = "black") +
    ggplot2::scale_x_continuous(name = "Quantile of standard\nnormal distribution", limits = theRange) +
    ggplot2::scale_y_continuous(name = "Quantile of\nstandardized residuals", limits = theRange) +
    ggplot2::coord_equal() +
    ggplot2::ggtitle("B.") +
    ggplot2::theme_bw() +
    ggplot2::theme(panel.grid.major = ggplot2::element_blank(),
                   panel.grid.minor = ggplot2::element_blank()) +
    ggplot2::theme(plot.title=ggplot2::element_text(hjust = 0,face = "bold", size = 10))


  tmp <- sp::spTransform(spatialData, sp::CRS(CRS_arg_utm))

  if(is.null(variogram_breaks)){
    variogram_breaks <- "default"
  } else {
    # variogram breaks must contain the distances (in kilometers) that specify
    # the variogram bins.
    # convert from kilometers to meters
    variogram_breaks <- variogram_breaks * 1e3
  }

  oVariogram <- geoR::variog(coords = sp::coordinates(tmp),
                             data = x$residuals$std_residual,
                             breaks = variogram_breaks)

  df2 <- tibble::tibble(distance = oVariogram$u / 1e3,
                       semivariogram = oVariogram$v,
                       stdDev = oVariogram$sd)

  p3 <- ggplot2::ggplot(df2) +
    ggplot2::geom_hline(yintercept = stats::var(df1$std_residual), color = "green") +
    ggplot2::geom_point(ggplot2::aes(x = distance, y = semivariogram)) +
    ggplot2::scale_x_continuous(name = "Distance, in kilometers") +
    ggplot2::scale_y_continuous(name = "Semivariogram") +
    ggplot2::expand_limits(y = 0) +
    ggplot2::ggtitle("C.") +
    ggplot2::theme_bw() +
    ggplot2::theme(panel.grid.major = ggplot2::element_blank(),
                   panel.grid.minor = ggplot2::element_blank()) +
    ggplot2::theme(plot.title=ggplot2::element_text(hjust = 0,face = "bold", size = 10))

  grid::grid.newpage()
  grid::pushViewport(grid::viewport(layout=grid::grid.layout(1,3)))
  print(p1, vp=grid::viewport(layout.pos.row=1, layout.pos.col=1))
  print(p2, vp=grid::viewport(layout.pos.row=1, layout.pos.col=2))
  print(p3, vp=grid::viewport(layout.pos.row=1, layout.pos.col=3))

}

#' @title Summarize the standardized residuals
#'
#' @description Print various statistics that summarize the standardized
#' residuals.
#'
#' @param object
#' Object of class CheckResiduals.
#'
#' @param ...
#' Required for compatibility with generic definition
#'
#' @export
#'
summary.CheckResiduals <- function(object, ...){

  theRange <- range(object$residuals$std_residual)

  cat( sprintf( "\n#########################################################################\n"))
  cat(sprintf("Summary of the standardized residuals\n"))
  cat(sprintf("Number = %d\n", length(object$residuals$std_residual)))
  cat(sprintf("Range: %g   %g\n", theRange[1], theRange[2]))
  cat(sprintf("Mean = %g\n", mean(object$residuals$std_residual)))
  cat(sprintf("Standard deviation = %g\n", stats::sd(object$residuals$std_residual)))
  cat(sprintf("Skewness = %g\n", e1071::skewness(object$residuals$std_residual)))
  cat(sprintf("Median = %g\n", stats::median(object$residuals$std_residual)))
  cat(sprintf("IQR = %g\n", stats::IQR(object$residuals$std_residual)))
  cat( sprintf( "#########################################################################\n"))
}

#' @title Plot arithmetic signs of the standardized residuals at the field
#' sample locations
#'
#' @description
#' If a standardized residual is greater than 0, then the symbol, which is
#' plotted at the location of the associated field sample, is red.
#' However, if a standardized residual is less than (or equal to) 0,
#' then the plot symbol is blue.
#' This resulting plot is an incomplete map.
#' To make the map useful, additional information
#' such as geographic boundaries should be added, and the map should be
#' projected properly. These tasks are accomplished using the functions in
#' library MapUtilities.
#'
#' @param object
#' Object of class CheckTransect.
#'
#' @param spatialData
#' Container with the spatial data and associated information.
#' See the User's Guide for a detailed description of this container.
#'
#' @param ...
#' Additional arguments that are passed to function
#' MappingUtilities::Add_Marked_Points.
#'
#' @return A ggplot2 object to implement the previously-described plotting.
#'
#' @export
#'
plotResidualMap <- function(object, spatialData, ...){

  df <- tibble::as_tibble(sp::coordinates(spatialData))
  df$arithmetic_sign <- object$residuals$arithmetic_sign

  ggplot2::ggplot() +
    MappingUtilities::Add_Marked_Points(df, "arithmetic_sign", "Arithmetic sign",
                                        isColorContinuous = FALSE, ...)
}

#' @title Constructor of class to check the residuals
#'
#' @description
#' This class is used to check the standardized residuals. A residual is defined
#' as the difference between a field sample value and the expected value
#' of the associated predicted measurement; the residual is standardized
#' by dividing it by the standard deviation of the predicted
#' measurement.
#' This function calculates these standardized residuals as well as
#' various related quantities that are needed to perform the
#' checks; the checks themselves are performed using other functions
#' that are associated with this class.
#'
#' @param spatialData
#' Container with the spatial data and associated information.
#' See the User's Guide for a detailed description of this container.
#'
#' @param rawSamples
#' S4 class returned by function rstan::sampling, which contains samples
#' of the posterior probability density function.
#'
#' @details
#' The expected value of a predicted measurement is the process mean;
#' the standard deviation of a predicted measurement is the
#' square root of its variance, which equals
#' the process variance plus the variance of the measurement error.
#'
#' @return A list with the following components is returned.
#' @return \item{residuals}{R-language tibble that contains
#' \itemize{
#'   \item \code{X}, the field samples values
#'   \item \code{Y_mean}, the process means
#'   \item \code{Y_sd}, the process standard deviations
#'   \item \code{tau}, the standard deviations of the predicted measurements
#'   \item \code{std_residual}, the standardized residuals
#'   \item \code{relation}, a character vector. If an element is "Greater", then
#'          the associated standardized residual is greater than 0. If an
#'          element is "Less", then the associated standardized residual is
#'          less than or equal to 0.
#' }
#' }
#' @return \item{call}{Function call.}
#'
#' @export
#'
CheckResiduals <- function(spatialData, rawSamples){

  postPdfSamples <- rstan::extract(rawSamples)

  df <- tibble::tibble(X = stanData$X,
                       Y_mean = apply(postPdfSamples$Y_mean, 2, mean),
                       Y_sd = apply(postPdfSamples$Y_sd, 2, mean),
                       tau = sqrt(Y_sd^2 + spatialData@data$me_sd^2),
                       std_residual = (X - Y_mean) / tau,
                       arithmetic_sign = ifelse(std_residual > 0, "Positive", "Negative"))

  rval = list(
    residuals = df,
    call = sys.call())

  class(rval) <- "CheckResiduals"

  return(rval)
}



