#' @title Plot sample locations along a transect
#'
#' @description
#' This plot
#' is an incomplete map. To make the map useful, additional information
#' such as geographic boundaries should be added, and the map should be
#' projected properly. These tasks are accomplished using the functions in
#' library MapUtilities.
#'
#' @param object
#' Object of class CheckTransect or class ViewTransect.
#'
#' @return A ggplot2 object to implement the previously-described plotting.
#'
#' @export
#'
plotSampleLocations <- function(object) {
  ggplot2::ggplot() +
    MappingUtilities::Add_GreatCircleArc(object$end_points, color = "red") +
    MappingUtilities::Add_Points(object$sampleLocations_longlat,
                                 shape = 16)
}

#' @title Graphical check of the predictions along the transect
#'
#' @description
#' The check involves four plots, In each plot, the plotted quantity is a
#' function of distance along the transect. The plotted quantities are
#' (1) the measured values of the field samples that are included in the
#' transect.
#' (2) quantile summaries of the predicted process mean at the field sample
#' locations.
#' (3) quantile summaries of the predicted process standard deviation at the
#' field sample locations
#' (4) quantile summaries of the predicted measurements at the
#' field sample locations, and the measured values of the field samples.
#'
#' @param x
#' Object of class CheckTransect.
#'
#' @param rawSamples
#' S4 class returned by function rstan::sampling, which contains samples
#' of various model parameters
#'
#' @param yAxisLabel
#' Label for the y axis
#'
#' @param pl
#' Lower probability for the quantile summaries.
#'
#' @param pu
#' Upper probability for the quantile summaries.
#'
#' @param ...
#' Required for compatibility with generic definition
#'
#' @details
#' Arguments \code{pl} and \code{pu} must be greater than 0 and less than 1.
#' Furthermore, \code{pl} must be less than \code{pu}.
#' Usually \code{pl} equals 1 - \code{pu}.
#'
#' @export
#'
plot.CheckTransect <- function(x, rawSamples,
                               yAxisLabel,
                               pl = 0.025, pu = 0.975, ...){

  CalcQuantile <- function(p, mu, sigma){
    return(mean(stats::qnorm(p, mean = mu, sd = sigma)))
  }

  areInTransect <- x$areInTransect
  N <- sum(areInTransect)

  postSamples <- rstan::extract(rawSamples)
  Y_mean <- postSamples$Y_mean[, areInTransect]
  Y_sd <- postSamples$Y_sd[, areInTransect]

  df <- tibble::tibble(distance = x$distanceAlongTransect,
                       X = x$Value,
                       Y_mean_ql = apply(Y_mean, 2, stats::quantile, probs = pl, names = FALSE),
                       Y_mean_q50 = apply(Y_mean, 2, stats::quantile, probs = 0.50, names = FALSE),
                       Y_mean_qu = apply(Y_mean, 2, stats::quantile, probs = pu, names = FALSE),
                       Y_sd_ql = apply(Y_sd, 2, stats::quantile, probs = pl, names = FALSE),
                       Y_sd_q50 = apply(Y_sd, 2, stats::quantile, probs = 0.50, names = FALSE),
                       Y_sd_qu = apply(Y_sd, 2, stats::quantile, probs = pu, names = FALSE),
                       X_pi_ql = vector(mode = "numeric", length = N),
                       X_pi_q50 = vector(mode = "numeric", length = N),
                       X_pi_qu = vector(mode = "numeric", length = N))

  for(i in 1:N){
    tau <- sqrt(Y_sd[, i] * Y_sd[, i] + x$me_sd[i]^2)
    df$X_pi_ql[i] <- CalcQuantile(pl, Y_mean[, i], tau)
    df$X_pi_q50[i] <- CalcQuantile(0.50, Y_mean[, i], tau)
    df$X_pi_qu[i] <- CalcQuantile(pu, Y_mean[, i], tau)
  }

  yLimits <- range(df[, c("X", "Y_mean_ql", "Y_mean_q50", "Y_mean_qu",
                          "X_pi_ql", "X_pi_q50", "X_pi_qu")])

  p1 <- ggplot2::ggplot(df) +
    ggplot2::geom_point(ggplot2::aes(x = distance, y = X), color = "red") +
    ggplot2::scale_x_continuous(name = "Distance, in kilometers") +
    ggplot2::scale_y_continuous(name = yAxisLabel, limits = yLimits) +
    ggplot2::ggtitle("A. Data") +
    ggplot2::theme_bw() +
    ggplot2::theme(panel.grid.major = ggplot2::element_blank(),
                   panel.grid.minor = ggplot2::element_blank()) +
    ggplot2::theme(plot.title=ggplot2::element_text(hjust = 0,face = "bold", size = 10))

  p2 <- ggplot2::ggplot(df) +
    ggplot2::geom_pointrange(ggplot2::aes(x = distance, y = Y_mean_q50,
                        ymin = Y_mean_ql, ymax = Y_mean_qu), fatten = 1) +
    ggplot2::scale_x_continuous(name = "Distance, in kilometers") +
    ggplot2::scale_y_continuous(name = yAxisLabel, limits = yLimits) +
    ggplot2::ggtitle("B. Process Mean") +
    ggplot2::theme_bw() +
    ggplot2::theme(panel.grid.major = ggplot2::element_blank(),
                   panel.grid.minor = ggplot2::element_blank()) +
    ggplot2::theme(plot.title=ggplot2::element_text(hjust = 0,face = "bold", size = 10))

  p3 <- ggplot2::ggplot(df) +
    ggplot2::geom_pointrange(ggplot2::aes(x = distance, y = Y_sd_q50,
                        ymin = Y_sd_ql, ymax = Y_sd_qu), fatten = 1) +
    ggplot2::scale_x_continuous(name = "Distance, in kilometers") +
    ggplot2::scale_y_continuous(name = yAxisLabel) +
    ggplot2::ggtitle("C. Process Standard Deviation") +
    ggplot2::theme_bw() +
    ggplot2::theme(panel.grid.major = ggplot2::element_blank(),
                   panel.grid.minor = ggplot2::element_blank()) +
    ggplot2::theme(plot.title=ggplot2::element_text(hjust = 0,face = "bold", size = 10))

  p4 <- ggplot2::ggplot(df) +
    ggplot2::geom_point(ggplot2::aes(x = distance, y = X), color = "red") +
    ggplot2::geom_pointrange(ggplot2::aes(x = distance, y = X_pi_q50,
                        ymin = X_pi_ql, ymax = X_pi_qu), fatten = 1) +
    ggplot2::scale_x_continuous(name = "Distance, in kilometers") +
    ggplot2::scale_y_continuous(name = yAxisLabel, limits = yLimits) +
    ggplot2::ggtitle("D. Predicted Measurements and Data") +
    ggplot2::theme_bw() +
    ggplot2::theme(panel.grid.major = ggplot2::element_blank(),
                   panel.grid.minor = ggplot2::element_blank()) +
    ggplot2::theme(plot.title=ggplot2::element_text(hjust = 0,face = "bold", size = 10))

  grid::grid.newpage()
  grid::pushViewport(grid::viewport(layout=grid::grid.layout(4,1)))
  print(p1, vp=grid::viewport(layout.pos.row=1, layout.pos.col=1))
  print(p2, vp=grid::viewport(layout.pos.row=2, layout.pos.col=1))
  print(p3, vp=grid::viewport(layout.pos.row=3, layout.pos.col=1))
  print(p4, vp=grid::viewport(layout.pos.row=4, layout.pos.col=1))

}


#' @title Graphical check of the statistical model using simulated data
#'
#' @description
#' The comparison involves 8 plots. For the first plot, the values of the
#' field samples that are included in the transect are plotted as a function
#' of distance along the transect.
#' For the remaining 7 plots, data are simulated at the
#' locations of the field samples that are included in the transect. The
#' simulated data are plotted as a function of distance along the transect
#'
#' @param object
#' Object of class CheckTransect.
#'
#' @param rawSamples
#' S4 class returned by function rstan::sampling, which contains samples
#' of various model parameters
#'
#' @param yAxisLabel
#' Character string that is a label for the y axis
#'
#' @param seed
#' Positive-value integer that is a random number seed.
#'
#' @details
#' The simulations involve random sampling from a normal probability density
#' function.
#' To make the simulations reproducible, the seed must be set to a
#' positive-valued integer.
#'
#' @export
#'
plotSimulatedTransects <- function(object, rawSamples,
                                   yAxisLabel,
                                   seed = 1L){

  simulate <- function(Y_mean, Y_sd, me_sd, seed){

    set.seed(seed)

    rs <- stats::rnorm(length(Y_mean), mean = Y_mean,
                sd = sqrt(Y_sd * Y_sd + me_sd^2))

    return(rs)
  }

  PlotTransectPoints <- function(df, Y, figLabel, yAxisLabel,
                                 point_color = "black",
                                 yLimits = NULL){
    p <- ggplot2::ggplot(df) +
      ggplot2::geom_point(ggplot2::aes_string(x = "distance", y = Y), color = point_color) +
      ggplot2::scale_x_continuous(name = "Distance, in kilometers") +
      ggplot2::scale_y_continuous(name = yAxisLabel,
                         limits = yLimits) +
      ggplot2::ggtitle(figLabel) +
      ggplot2::theme_bw() +
      ggplot2::theme(panel.grid.major = ggplot2::element_blank(),
                     panel.grid.minor = ggplot2::element_blank()) +
      ggplot2::theme(plot.title=ggplot2::element_text(hjust = 0,face = "bold", size = 10))

    return(p)
  }

  nSimulations <- 7

  postSamples <- rstan::extract(rawSamples)

  areInTransect <- object$areInTransect
  N <- sum(areInTransect)

  Y_mean <- postSamples$Y_mean[, areInTransect]
  Y_sd <- postSamples$Y_sd[, areInTransect]
  me_sd <- object$me_sd

  df <- tibble::tibble(distance = object$distanceAlongTransect,
                       X = object$Value)

  for(i in 1:nSimulations){
    cmd <- paste0("df$S", i, " <- ", "simulate(Y_mean[", i, ", ], Y_sd[", i, ", ], me_sd[", i, "], seed + ", i, ")")
    eval(parse(text = cmd))
  }

  yLimits <- range(df[, -1])

  p1 <- PlotTransectPoints(df, "X", "A. Measured", yAxisLabel = yAxisLabel,
                           yLimits = yLimits, point_color = "red")

  for(i in 1:nSimulations){
    figLabel <- paste0(LETTERS[i+1], ". Simulated")
    cmd <- paste0("p", as.character(i+1),
                  "<- PlotTransectPoints(df, 'S", as.character(i), "',",
                  paste0("'", figLabel, "'"),
                  ", yAxisLabel = yAxisLabel, yLimits = yLimits)")
    eval(parse(text = cmd))
  }

  grid::grid.newpage()
  grid::pushViewport(grid::viewport(layout=grid::grid.layout(4,2)))
  plot(p1, vp=grid::viewport(layout.pos.row=1, layout.pos.col=1))
  plot(p2, vp=grid::viewport(layout.pos.row=2, layout.pos.col=1))
  plot(p3, vp=grid::viewport(layout.pos.row=3, layout.pos.col=1))
  plot(p4, vp=grid::viewport(layout.pos.row=4, layout.pos.col=1))
  plot(p5, vp=grid::viewport(layout.pos.row=1, layout.pos.col=2))
  plot(p6, vp=grid::viewport(layout.pos.row=2, layout.pos.col=2))
  plot(p7, vp=grid::viewport(layout.pos.row=3, layout.pos.col=2))
  plot(p8, vp=grid::viewport(layout.pos.row=4, layout.pos.col=2))
}

# CalcTransect
#
# Determine which samples are projected into the transect, calculate their
# distances along the transect and perpendicular to the transect
#
# sampleLocations_longlat: Dataframe containing the locations of the field samples. The
#     locations are specified by their longitude and latitude, for which the
#     column headings in the dataframe are "long" and "lat".
# transect_endpoints: Dataframe containing the locations of the endpoints of the
#     transect. The locations are specified by their longitude and latitude,
#     for which the column headings in the dataframe are "long" and "lat".
# maxTransectOffset: The maximum distance between the transect and a sample
#     that is projected into the transect. The units are meters.
# CRS_arg_longlat: arguments to function sp::CRS specfying the projection
#              for the argument sampleLocations_longlat
# CRS_arg_utm: arguments to function sp::CRS specifying the projection
#              for the utm coordinates
#
# The function value is a list containing vectors areInTransect,
# distanceAlongTransect, and distanceFromTransect. These three vectors relate to
# dataframe sampleLocations_longlat: Element i in a vector pertains to row (sample) i in
# sampleLocations_longlat. Element i in areInTransect, which is a logical value, indicates
# whether sample i is projected into the transect. If element i in areInTransect
# is TRUE, then element i in distanceAlongTransect is the distance of the
# the sample along the transect. Otherwise, element i in distanceAlongTransect
# is NA. Similarly, if element i in areInTransect
# is TRUE, then element i in distanceFromTransect is the distance of the
# the sample perpendicular to the transect. Otherwise, element i in
# distanceFromTransect is NA.
#
CalcTransect <- function(sampleLocations_longlat,
                         maxTransectOffset,
                         transect_endpoints,
                         CRS_arg_longlat,
                         CRS_arg_utm){

  CalcUtmCoords <- function(df, CRS_arg_longlat, CRS_arg_utm) {

    tmp <- sp::SpatialPoints(df[, c("long","lat"), drop = FALSE],
                         proj4string=sp::CRS(CRS_arg_longlat))
    tmp.utm <- sp::spTransform(tmp, sp::CRS(CRS_arg_utm) )
    coords <- sp::coordinates(tmp.utm)
    colnames(coords) <- c("Easting", "Northing")
    return(coords)
  }


  rotate <- function(x, angle) {
    return(c(x[1] * cos(angle) + x[2] * sin(angle),
             -x[1] * sin(angle) + x[2] * cos(angle)))
  }

  # calculate the utm coordinates of the locations of the sample data
  sample_coords <- CalcUtmCoords(sampleLocations_longlat, CRS_arg_longlat, CRS_arg_utm)

  # calculate the utm coordinates of the locations of the transect endpoints
  endpts_coords <- CalcUtmCoords(transect_endpoints, CRS_arg_longlat, CRS_arg_utm)

  # translate the coordinates --- the new origin is the first endpoint
  # of the transect
  origin <- endpts_coords[1, ]

  t_endpts_coords <- t(apply(endpts_coords, 1, function(x){x - origin}))
  t_sample_coords <- t(apply(sample_coords, 1, function(x){x - origin}))

  # rotate the coordinate system so that the positive x axis passes
  # through the second endpoint of the transect
  angle <- atan2(t_endpts_coords[2, 2], t_endpts_coords[2, 1])

  tr_endpts_coords <- t(apply(t_endpts_coords, 1, rotate, angle))
  tr_sample_coords <- t(apply(t_sample_coords, 1, rotate, angle))

  colnames(tr_endpts_coords) <- c("x", "y")
  colnames(tr_sample_coords) <- c("x", "y")

  # in tr_endpts_coords and tr_sample_coords,
  # x is the distance along the transect.
  # y is the distance from (perpendicular to) the transect

  # If a sample is projected into the transect, then it must satisfy
  # two criteria:
  # (1) It must be within distance maxTransectOffset of the transect.
  # (2) Its projection must be between the two endpoints of the transect.
  areInTransect <- abs(tr_sample_coords[, "y"]) < maxTransectOffset &
    tr_sample_coords[, "x"] >= 0 &
    tr_sample_coords[, "x"] <= tr_endpts_coords[2, "x"]

  distanceAlongTransect = tr_sample_coords[, "x"]
  distanceAlongTransect[!areInTransect] <- NA

  distanceFromTransect = abs(tr_sample_coords[, "y"])
  distanceFromTransect[!areInTransect] <- NA

  return(list(areInTransect = areInTransect,
              distanceAlongTransect = distanceAlongTransect,
              distanceFromTransect = distanceFromTransect))

}


#' @title Constructor of class to check the model along a transect
#'
#' @description
#' This constructor calculates various quantities are needed to
#' to check the suitability of the statistical model along transect.
#' The checks themselves are performed using graphics, which are plotted
#' by other functions associated with this class.
#'
#' @param spatialData
#' Container with the spatial data and associated information.
#' See the User's Guide for a detailed description of this container.
#'
#' @param maxOffset
#' Real-valued scalar that specifies the maximum offset---namely, the
#' maximum distance (in kilometers) between a field sample and the transect.
#' Field
#' samples for which the offset is less than maxOffset are included in the
#' transect; otherwise, they are not.
#'
#' @param end_points
#' R-language dataframe containing the locations of the two ends of the
#' transect. Each row pertains one end.
#' The locations are specified with longitude and latitude; the column headings
#' in the dataframe are "long" and "lat". The datum for the longitudes and
#' latitudes must be the same as that used for argument \code{spatialData}.
#'
#' @param CRS_arg_utm
#' Character string containing the arguments to function sp::CRS that
#' specify the projection, in universal transverse mercator (utm) coordinates,
#' for arguments \code{spatialData} and \code{end_points}.
#'
#' @return A list with the following components is returned.
#' @return \item{maxOffset}{Argument \code{maxOffset}.}
#' @return \item{end_points}{Argument \code{end_points}.}
#' @return \item{CRS_arg_longlat}{Character string containing the arguments to
#' function sp::CRS that specify the projection for longitudes and latitudes
#' for arguments \code{spatialData} and \code{end_points.}}
#' @return \item{CRS_arg_utm}{Argument \code{CRS_arg_utm}.}
#' @return \item{areInTransect}{Logical vector indicating whether the field
#' samples in \code{spatialData} are included in the transect.}
#' @return \item{distanceAlongTransect}{Real-valued vector containing the
#' distances along the transect of just the field samples that are in the
#' transect.}
#' @return \item{distanceFromTransect}{Real-valued vector containing the
#' distances from the transect (namely, the offset perpendicular to the
#' transect) of just the field samples that are in the transect.}
#' @return \item{sampleLocations_longlat}{R-language tibble containing the
#' longitudes and latitudes of just the field samples that are included in
#' the transect.}
#' @return \item{Value}{Real-valued vector containing the measured values of
#' just the field samples that are in the transect.}
#' @return \item{call}{Function call.}
#'
#' @export
#'
CheckTransect <- function(spatialData, maxOffset, end_points,
                          CRS_arg_utm){

  sampleLocations_longlat <- sp::coordinates(spatialData)
  CRS_arg_longlat <- sp::proj4string(spatialData)

  tmp <- CalcTransect(sampleLocations_longlat,
                      maxOffset * 1e3,
                      end_points,
                      CRS_arg_longlat,
                      CRS_arg_utm)

  rval <- list(maxOffset = maxOffset,
               end_points = end_points,
               CRS_arg_longlat = CRS_arg_longlat,
               CRS_arg_utm = CRS_arg_utm,
               areInTransect = tmp$areInTransect,
               distanceAlongTransect = tmp$distanceAlongTransect[tmp$areInTransect] / 1e3,
               distanceFromTransect = tmp$distanceFromTransect[tmp$areInTransect] /  1e3,
               sampleLocations_longlat = tibble::as_tibble(sampleLocations_longlat[tmp$areInTransect, ]),
               Value = spatialData$Value[tmp$areInTransect],
               me_sd = spatialData$me_sd[tmp$areInTransect],
               call = sys.call())

  class(rval) <- "CheckTransect"

  return(rval)

}

