#' @title Plot the results of the cross validation test
#'
#' @description There are two plots: One plot is the deviances versus distance
#' between basis function centers, and the other plot is mean squared error
#' versus distance between basis function centers. At each distance, the
#' deviances for the \code{nFold} folds are summarized by the mean and the
#' standard error in the mean.
#' The mean squared errors are likewise summarized.
#'
#' @param x
#' Object of class CrossValidation
#'
#' @param jitter_width
#' Real-valued scalar specifying the spread of the random noise. See Details.
#'
#' @param ...
#' Required for compatibility with generic definition
#'
#' @details
#' At specified distance, the gray dots represent the measures of predictive
#' accuracy (that is, deviance or mean squared error). A small amount of random
#' noise has been added to the distance, which minimizes dots plotting atop one
#' another. The red dot represents the mean of the measures of predictive
#' accuracy. The top of the red vertical line corresponds
#' to one standand error above the mean; similarly, the bottom of the red
#' vertical line corresponds to one standard error below the mean.
#'
#' @return A ggplot2 object to implement the previously-described plotting.
#'
#' @export
#'
plot.CrossValidation <- function(x, jitter_width = 0.005, ...) {

  df <- tibble::tibble()
  for(i in 1:length(x$results)){
    df_tmp <- x$results[[i]]
    df_tmp$distance <- x$center_spacings[i]
    df <- rbind(df, df_tmp)
  }

  Mean_CL <- function(z){
    sigma <- sd(z) / sqrt(length(z))
    tibble::tibble(y = mean(z),
           ymin = y - sigma,
           ymax = y + sigma)
  }

  p1 <- ggplot2::ggplot(data = df, mapping = ggplot2::aes(x = distance, y = deviance)) +
    ggplot2::geom_jitter(height = 0, width = jitter_width, shape = 16, color = "gray65") +
    ggplot2::stat_summary(geom = "pointrange", fun.data = Mean_CL, color = "red") +
    ggplot2::scale_x_continuous(name = "Distance between basis function centers, in kilometers",
                                trans = "log10",
                                breaks = sort(unique(df$distance))) +
    ggplot2::scale_y_continuous(name = "Deviance") +
    ggplot2::ggtitle("A.") +
    ggplot2::theme_bw() +
    ggplot2::theme(plot.title = ggplot2::element_text(hjust = 0,face = "bold", size = 10))


  p2 <- ggplot2::ggplot(data = df, mapping = ggplot2::aes(x = distance, y = mse)) +
    ggplot2::geom_jitter(height = 0, width = jitter_width, shape = 16, color = "gray65") +
    ggplot2::stat_summary(geom = "pointrange", fun.data = Mean_CL, color = "red") +
    ggplot2::scale_x_continuous(name = "Distance between basis function centers, in kilometers",
                                trans = "log10",
                                breaks = sort(unique(df$distance))) +
    ggplot2::scale_y_continuous(name = "Mean squared error") +
    ggplot2::ggtitle("B.") +
    ggplot2::theme_bw() +
    ggplot2::theme(plot.title = ggplot2::element_text(hjust = 0,face = "bold", size = 10))

  grid::grid.newpage()
  grid::pushViewport(grid::viewport(layout=grid::grid.layout(2,1)))
  plot(p1, vp=grid::viewport(layout.pos.row=1, layout.pos.col=1))
  plot(p2, vp=grid::viewport(layout.pos.row=2, layout.pos.col=1))
}

#' @title Summarize the results of the cross validation test
#'
#' @description Construct a table that summarizes, at each of the spacings,
#' the two measures of predictive accuracy: the mean deviance with its
#' standard error, and
#' the mean of the mean squared error (MSE) with its standard error.
#'
#' @param object
#' Object of class CrossValidation
#'
#' @param ...
#' Required for compatibility with generic definition
#'
#' @export
#'
summary.CrossValidation <- function(object, ...) {

  cat( sprintf( "\n#########################################################################\n"))
  cat(sprintf("Summary\n"))
  cat(sprintf("The numbers in parentheses are the standard errors.\n\n"))
  cat( sprintf("             Mean                               Mean\n"))
  cat( sprintf("Spacing      Deviance                           MSE \n"))
  for(i in 1:length(object$results)){
    cat(sprintf("%-10d   %-10.3g (%-10.3g)            %-10.3g (%-10.3g)\n",
                object$center_spacings[i],
                mean(object$results[[i]]$deviance),
                stats::sd(object$results[[i]]$deviance),
                mean(object$results[[i]]$mse),
                stats::sd(object$results[[i]]$mse)))
  }
  cat( sprintf( "#########################################################################\n"))
}


# CalcPredictiveAccuracy
#
# Calculate the predictive accuracy for the specified basis functions and
# the test samples.
#
# areTestSamples: Logical vector that specifies which elements of vector X are
#                 test samples. The other elements of vector X are training
#                 samples.
# basis_functions: List that is returned by function Calc_basis_func, which has
#                 the complete documetantion of this list. The only list element
#                 that is
#                 pertinent to this function is called "basisFuncs", which
#                 is a matrix that contains the values of the basis functions
#                  that are associated with the elements of X.
# X: Vector containing the observed values of the field.
# adjMatrix: List that is returned by function CalcAdjacencyMatrix, which has
#                 has the complete documentation for this list.
# me_sd: Standard deviation of the measurement error.
# areLeftCensored: Logical vector that indicates which elements of vector X are
#                 left censored. If a element of X is left censored, then the
#                 corresponding element of vector areLeftCensored is set to
#                 TRUE. Otherwise, it is set to FALSE.
# BMNUS_sm: The compiled stan program BMNUS.
#
# Details
#
# Matrix basisFuncs is sparse; its elements are stored in "dictionary-of-keys"
# format.
#
# In the current implementation of the this function, a few of the stan
# sampling parameters are hard-wired. Although the chosen, hard-wired values
# seem to work in all cases, it may be prudent to set the values in the
# in the function call.
#
# The function value is a list containing three measures of predictive
# accuracy: the deviance, the mean squared error, and the weighted mean
# square error.
#
# Reference:
# Gelman, A., Carlin, J.B., Stern, H.S., Dunson, D.B., Vehtari, A., and
# Rubin, D.B., 2014, Bayesian data analysis (3d ed.): Boca Raton, Fla.,
# CRC Press, section 7.1 (p. 166-178).
CalcPredictiveAccuracy <- function(areTestSamples, spatialData, basis_functions,
                                   carQuantities,
                                   BMNUS_sm){

  require(Matrix, quietly = TRUE)
  require(rstan, quietly = TRUE)

  areTrainingSamples <- !areTestSamples

  # Transform matrix basisFuncs from "dictionary-of-keys" format  to
  # "compressed-row-storage" format.
  csrBasisFuncs <- as(basis_functions$basisFuncs[areTrainingSamples, ],
                      "RsparseMatrix")

  stanData <- list( N = sum(areTrainingSamples),
                    M = ncol(csrBasisFuncs),
                    UU = csrBasisFuncs@p + 1,
                    VV = csrBasisFuncs@j + 1,
                    WW = csrBasisFuncs@x,
                    nNeighborPairs = carQuantities$nNeighborPairs,
                    W_sparse = carQuantities$W_sparse,
                    D_sparse = carQuantities$D_sparse,
                    lambda = carQuantities$lambda,
                    X = spatialData@data$Value[areTrainingSamples],
                    me_sd = spatialData@data$me_sd[areTrainingSamples],
                    areLeftCensored = as.integer(spatialData@data$IndValue[areTrainingSamples] == "left"),
                    betaPar = c(2.5, 1.2),
                    gammaPar = c(2, 0.3),
                    cauchyPar = 3)

  # Function gen_int initializes some variables for the sampling
  gen_init <- function(){
    phi1 <- stats::rnorm(stanData$M, mean = 0.0, sd = 1e-6)
    phi2 <- stats::rnorm(stanData$M, mean = 0.0, sd = 1e-6)
    return(list(phi1 = phi1, phi2 = phi2))
  }

  # Sample the posterior pdf
  rawSamples <- rstan::sampling(BMNUS_sm, data = stanData, chains = 1,
                         iter = 2500, warmup = 500,
                         refresh = 500,
                         init = gen_init,
                         pars=c("rho", "phi1", "phi2"))

  # compute the three measures of predictive accuracy.
  B_test <- basis_functions$basisFuncs[areTestSamples, ]
  X_test <- spatialData@data$Value[areTestSamples]
  areLeftCensored_test <- as.integer(spatialData@data$IndValue[areTestSamples] == "left")
  me_sd_test <- spatialData@data$me_sd[areTestSamples]

  postPdfSamples <- rstan::extract(rawSamples)

  # There are two ways to compute the process mean (Y_mean) and the process
  # standard devation (Y_sd). Procedure A computes the averages of the random
  # variables phi1, phi2, and rho. Then these averages are used to compute
  # Y_mean and Y_sd. Procedure B compute Y_mean and Y_sd for each realization
  # of random variables phi1, phi2, and rho. Then it averages Y_mean and Y_sd
  # to obtain final values for Y_mean and Y_sd.
  #
  # The two procedures do not yield the same values. I think that procedure B
  # yields the most accurate measures of predictive accuracy.

  # Procedure A
  # Y_mean <- mean(stanData$X) + B_test %*% colMeans(postPdfSamples$phi1)
  # Y_sd <- mean(postPdfSamples$rho) * exp(B_test %*% colMeans(postPdfSamples$phi2))

  # Procedure B
  tmp1 <- Matrix::tcrossprod(postPdfSamples$phi1, B_test)
  Y_mean = mean(stanData$X) + Matrix::colMeans(tmp1)

  tmp2 <- exp(Matrix::tcrossprod(postPdfSamples$phi2, B_test))
  tmp3 <- Matrix::Diagonal(x = postPdfSamples$rho) %*% tmp2
  Y_sd = Matrix::colMeans(tmp3)

  logLikelihood <- 0.0
  squared_error <- numeric(length(X_test))
  weighted_squared_error <- numeric(length(X_test))
  for(j in 1:length(X_test)){
    sigma <- sqrt(Y_sd[j]^2 + (me_sd_test[j])^2)
    if(areLeftCensored_test[j] == 1){
      logLikelihood <- logLikelihood +
        stats::pnorm(X_test[j], mean = Y_mean[j], sd = sigma, log = TRUE)
    } else {
      logLikelihood <- logLikelihood +
        stats::dnorm(X_test[j], mean = Y_mean[j], sd = sigma, log = TRUE)
    }
    squared_error[j] <- (X_test[j] - Y_mean[j])^2
    weighted_squared_error[j] <- (X_test[j] - Y_mean[j])^2 / sigma^2
  }

  return(list(
    deviance = -2.0 * logLikelihood,
    mse = mean(squared_error),
    weighted_mse = mean(weighted_squared_error)))
}


# pXV
#
# Perform cross validation using parallel computation.
#
# spatialData
# Container with the spatial data and associated information.
# See the User's Guide for a detailed description of this container.
#
# areTestSamples: Logical vector that specifies which elements of vector X are
#                 test samples. The other elements of vector X are training
#                 samples.
# basis_functions: List that is returned by function Calc_basis_func, which has
#                 the complete documetantion of this list.
# nFolds: The number of folds in the cross validation.
# nCpuCores: The number of cpu cores that are used for the parallel computations.
# BMNUS_sm: The compiled stan program BMNUS.
#
# The function value is a list with nFold elements. Each element is another
# list---this list is returned from function CalcPredictiveAccuracy and
# contains three measures of predictive
# accuracy: the deviance, the mean squared error, and the weighted mean
# square error.
#
pXV <- function(spatialData,
                basis_functions, nFolds, nCpuCores, BMNUS_sm){

  CalcTestMatrix <- function(N, K = 10L, seed = NULL){

    # N must be an integer
    stopifnot(stringi::stri_cmp_eq(typeof(N), "integer"))

    # We want M1 groups of size N1 and M2 groups of size N2
    # The total number of groups is K. So, M1 + M2 = K
    N1 <- as.integer(N / K)
    N2 <- N1 + 1
    M1 <- N2 * K - N
    M2 <- K - M1

    # Check
    stopifnot(N == M1 * N1 + M2 * N2)

    set.seed(seed)
    indices <- sample.int(N)
    areTest <- matrix(FALSE, nrow = N, ncol = K)
    count <- 1
    for(k in 1:K){
      if(k <= M1){
        itmp <- indices[count:(count+N1-1)]
        areTest[itmp, k] <- TRUE
        count <- count + N1
      } else {
        itmp <- indices[count:(count+N2-1)]
        areTest[itmp, k] <- TRUE
        count <- count + N2
      }
    }

    # Check that each observation is in 1, and only 1, group
    stopifnot(rowSums(areTest) == 1)

    # Check that the groups are setup as specified above
    stopifnot(colSums(areTest) == c(rep.int(N1, M1), rep.int(N2, M2)))

    return(areTest)

  }

  carQuantities <- SparseCarQuantities(basis_functions)

  testMatrix <- CalcTestMatrix(nrow(spatialData), K = nFolds)

  CL <- parallel::makeCluster(nCpuCores)

  result <- parallel::parCapply(CL, testMatrix, CalcPredictiveAccuracy,
                                spatialData, basis_functions, carQuantities,
                                BMNUS_sm)

  parallel::stopCluster(CL)

  # Extract the deviance, mse, wmse
  N <- length(result)
  df <- tibble::tibble(deviance = numeric(N),
                       mse = numeric(N),
                       wmse = numeric(N))
  for(n in 1:N){
    df[n, 1:3] <- unlist(result[[n]])
  }

  # do not return the weighted mean square error
  return(df[, -3])
}

# Find the samples that are within a specified region
#
# spatialData
# Container with the spatial data and associated information.
# See the User's Guide for a detailed description of this container.
#
# region
# R-language data frame listing the longitudes and latitudes of points
# along the boundary the region. Within the data frame, longitudes are
# called "long", and latitudes are called "lat".
#
# The return value is a logical vector, for which the length equals the
# the number of samples. A vector element is TRUE if the sample is within
# the region; otherwise, it is FALSE.
#
FindSamplesInsideRegion <- function(spatialData, region){
  sample_coordinates <- sp::coordinates(spatialData)

  indicator <- sp::point.in.polygon(sample_coordinates[, "long"],
                                    sample_coordinates[, "lat"],
                                    region$long,
                                    region$lat)
  return(indicator == 1)
}

# Subset container spatial data
#
# spatialData
# Container with the spatial data and associated information.
# See the User's Guide for a detailed description of this container.
#
# isSampleInsideRegion
# Logical vector return by function FindSamplesInsideRegion
#
# The return value is a container of type spatial data, but the container has
# only those samples that are within the previously-specified region.
#
# SubsetFieldSamples <- function(spatialData, isSampleInsideRegion){
#   return(spatialData[isSampleInsideRegion, ])
# }

# Subset previously-computed basis fuctions
#
# basis_functions
# Container with the basis functions at a specified spacing (See the
# documentation for the class BasisFunctions.)
#
# isSampleInsideRegion
# Logical vector return by function FindSamplesInsideRegion
#
# region
# R-language data frame listing the longitudes and latitudes of points
# along the boundary the region. Within the data frame, longitudes are
# called "long", and latitudes are called "lat".
#
# The return value is a container of type basis_functions, but the container has
# only those basis functions that are within the previously-specified region.
#
SubsetBasisFunctions <- function(basis_functions,
                                 isSampleInsideRegion,
                                 region){
  require(Matrix, quietly = TRUE)

  indicator <- sp::point.in.polygon(basis_functions$centers_longlat[, "long"],
                                    basis_functions$centers_longlat[, "lat"],
                                    region$long,
                                    region$lat)
  isBfInsideRegion <- indicator == 1

  # Sometimes, a basis function has no neighbors, which is
  # manifest as a row of zeros in the adjacency matrix. This basis function
  # must be eliminated. Equivalently,
  # only those basis functions are kept that have 1 or more neighbors.
  tmp1 <- list(sepDistance = basis_functions$sepDistance,
               basisFuncs = basis_functions$basisFuncs[isSampleInsideRegion, isBfInsideRegion],
               centers_utm = basis_functions$centers_utm[isBfInsideRegion, ],
               centers_longlat = basis_functions$centers_longlat[isBfInsideRegion, ])
  result <- SparseCarQuantities(tmp1)
  if(typeof(result) == "integer"){
    tmp2 <- list(sepDistance = tmp1$sepDistance,
                 basisFuncs = tmp1$basisFuncs[, -result],
                 centers_utm = tmp1$centers_utm[-result, ],
                 centers_longlat = tmp1$centers_longlat[-result, ])

  } else{
    tmp2 <- tmp1
  }

  return(tmp2)
}

#' @title Constructor of class to perform cross validation test
#'
#' @description
#' This constructor performs a cross validation test for one or more spacings
#' between basis function centers. The test may be performed either for
#' the entire domain or for a sub-domain.
#' The test is performed using parallel processing to reduce computation time.
#'
#' @param spatialData
#' Container with the spatial data and associated information.
#' See the User's Guide for a detailed description of this container.
#'
#' @param BMNUS_sm
#' The compiled stan program BMNUS.
#'
#' @param oBasisFunctions
#' Object of class BasisFunctions.
#'
#' @param nFolds
#' Number of folds for the cross validation test.
#'
#' @param nCpuCores
#' Number of processor (cpu) cores that are used for the parallel calculations.
#' This parameter should be less than the number of logical cores.
#'
#' @param selected_spacings
#' Character vector that contains the spacings for which the cross validation
#' is performed. That is, sometimes the cross
#' validation test needs to be perfromed for only some of the spacings that
#' were previously computed by class \code{BasisFunctions}. If so, these
#' spacings are specified in this vector.
#'
#' @param directory_name
#' Character string containing the name of the directory that will store
#' the files with the results of the cross validation test.
#'
#' @param sub_domain_boundary
#' R-language tibble that contains the longitudes and latitudes of the
#' points that define the boundary of a sub-domain.
#' Within the tibble, latitude and longitude are designated "lat" and "long".
#' This sub-domain must be a subset of the domain.
#'
#' @details
#' The duration of the computations depends upon many factors, including
#' the number of basis functions, the number of measurements in the domain,
#' the number of folds, and the speed of the computer. Thus, predicting the
#' duration is practically impossible. Nonetheless, the duration is much longer
#' than the other computations in package BMNUS.
#'
#' Sometimes, it may desirable to perform a cross validation test for a portion
#' of a domain. For example, if the domain has areas with different sampling
#' densities, then it may desirable to perform tests for individual areas.
#' This can be accomplished by using argument \code{sub_domain_boundary}---the
#' cross validation test is performed only for the samples within the specified
#' sub-domain.
#'
#' @return A list with the following components is returned.
#' @return \item{nFolds}{Argument \code{nFolds}.}
#' @return \item{nCpuCores}{Argument \code{nCpuCores}.}
#' @return \item{selected_spacings}{Argument \code{selected_spacings}.}
#' @return \item{center_spacings}{Spacings between basis function centers,
#' in kilometers, at which the cross validation test is performed.}
#' @return \item{results}{R-language list, for which the number of elements
#' equals the  number of spacings between basis function centers.
#' That is, each element of \code{results} is associated with an element of
#' \code{center_spacings}.}
#' Each element of \code{results} is a tibble
#' that contains the two measures of predictive accuracy: the deviance, and
#' the mean squared error (Gelman and others, 2014, p. 166-178). The number
#' of rows in the tibble equals the number of folds.
#' @return \item{call}{Function call.}
#'
#' @references
#' Gelman, A., Carlin, J.B., Stern, H.S., Dunson, D.B., Vehtari, A., and
#' Rubin, D.B., 2014, Bayesian data analysis (3d ed.): Boca Raton, Fla.,
#' CRC Press.
#'
#' @export
#'
CrossValidation <- function(spatialData, BMNUS_sm,
                            oBasisFunctions,
                            nFolds, nCpuCores,
                            selected_spacings = NULL,
                            directory_name = paste0("CrossValidation_",nFolds,"-Folds"),
                            sub_domain_boundary = NULL){

  if(!dir.exists(directory_name)){
    dir.create(directory_name)
  }

  if(is.null(selected_spacings)){
    indices <- 1:length(oBasisFunctions$spacings)
  } else {
    indices <- which(oBasisFunctions$spacings %in% sort(selected_spacings))
  }

  filenames <- paste0(directory_name, "\\XV_", oBasisFunctions$spacings[indices], ".csv")

  if(!is.null(sub_domain_boundary)){
    # Subset the data
    isSampleInsideRegion <- FindSamplesInsideRegion(spatialData, sub_domain_boundary)
    spatialData <- spatialData[isSampleInsideRegion, ]
  }

  N <- length(indices)
  results <- list(N)

  for(i in 1:N){

    load(oBasisFunctions$file_info$filenames[indices[i]])

    if(is.null(sub_domain_boundary)){
      results[[i]] <- pXV(spatialData, basis_functions, nFolds, nCpuCores, BMNUS_sm)
    } else {
      # Subset the basis functions
      basis_functions_sub <- SubsetBasisFunctions(basis_functions, isSampleInsideRegion, sub_domain_boundary)
      results[[i]] <- pXV(spatialData, basis_functions_sub, nFolds, nCpuCores, BMNUS_sm)
    }

    readr::write_csv(results[[i]], filenames[i])

    cat(sprintf("Finished spacing %g\n", oBasisFunctions$file_info$center_spacings[indices[i]]))
  }

  rval <- list(nFolds = nFolds,
               nCpuCores = nCpuCores,
               selected_spacings = selected_spacings,
               center_spacings = oBasisFunctions$file_info$center_spacings[indices],
               results = results,
               call = sys.call())

  class(rval) <- "CrossValidation"

  return(rval)
}

