#' @title Calculate quantities needed for the sparse CAR model
#'
#' @description
#' The sparse CAR model, which was developed by M. Joseph, does not
#' directly require the adjacency matrix. Rather, it requires a scalar,
#' two vectors, and a matrix that are derived from the adjacency matrix.
#' These quantities are calculated before calling stan program BMNUS, so
#' that the quantities are not calculated for each chain, which would be
#' repetitous. These quantities are calculated in this function.
#'
#' @param basis_functions
#' R-language list that is generated by class BasisFunctions. This list is
#' described in the documentation for function \code{BasisFunctions}.
#'
#' @param tol
#' Positive real number. See details.
#'
#' @details
#' Calculating the adjacency matrix involves determining the neighbors of every
#' basis function center. This determination is based on the distance between
#' the neighbors and
#' is controlled by argument \code{tol}. That is, if a (normalized) distance
#' is between \code{1 - tol} and \code{1 + tol}, then the associated center is
#' deemed a neighbor. Otherwise, it is not.
#'
#' @return If the adjacency matrix has one or more rows that are comprised
#' of just zeros, then the indices of those rows are returned as an
#' integer-valued vector. Otherwise, a list with the following components
#' is returned.
#' @return \item{nNeighborPairs}{Integer-valued scalar. Number of neighbor
#' pairs in the adjacency matrix.}
#' @return \item{W_sparse}{Integer-valued matrix. The first column lists the row
#' indices of the non-zero elements in the upper triangle of the adjacency
#' matrix. Only the upper triangle is used because the adjacency matrix is
#' symmetric. The second column lists the associated column indices of the
#' adjacency matrix.}
#' @return \item{D_sparse}{Real-valued vector. This vector is the diagonal of
#' diagonal matrix D that is used in sparse CAR formulation. See the
#' reference.}
#' @return \item{lambda}{Real-valued vector. This vector comprises the
#' eigenvalues of matrix D^(-1/2) W D^(-1/2) that is used in sparse CAR
#' formulation. See the reference.}
#'
#' @references
#' Joseph, Max, Exact sparse CAR models in Stan, Internet web page available at
#' http://mc-stan.org/documentation/case-studies/mbjoseph-CARStan.html,
#' last accessed September 2019.
#'
#' @export
#'
SparseCarQuantities <- function(basis_functions, tol = 1e-6){

  require(Matrix)

  CalcDist <- function( coords1, coords2 ) {
    dx <- coords1[1] - coords2[, 1]
    dy <- coords1[2] - coords2[, 2]
    return( sqrt( dx^2 + dy^2 ) )
  }

  # Calculate the indices of the non-zero elements in the adjacency matrix
  CalcMatrixIndices <- function(centers_coord, lb, ub, scale){

    # Number of rows and columns in the adjacency matrix
    N <- nrow(centers_coord)

    # estNNZ is an estimated number of non-zero entries in the
    # adjacency matrix
    estNNZ <- as.integer(scale * N * N^(1/3) + ifelse(N <= 200, N^2, 200^2))

    # row and column indices for the non-zero elements of the adjacency matrix
    row_indices <- integer(estNNZ)
    col_indices <- integer(estNNZ)

    # number of non-zero elements in the adjacency matrix
    NNZ <- 0

    error_condition <- 0L

    # indices of rows that comprise just 0s
    zero_rows <- integer(0)
    for(i in 1:N){
      distances <- CalcDist(centers_coord[i, ], centers_coord)
      indices <- which(lb < distances & distances < ub)
      M <- length(indices)
      if(NNZ + M > estNNZ){
        error_condition <- 1L
        break
      }
      if(M > 0){
        row_indices[(NNZ+1):(NNZ + M)] <- i
        col_indices[(NNZ+1):(NNZ + M)] <- indices
        NNZ <- NNZ + M
      } else {
        error_condition <- 2L
        zero_rows <- c(zero_rows, i)
      }
    }

    # Remove the unused elements at the end of these vectors.
    row_indices <- row_indices[1:NNZ]
    col_indices <- col_indices[1:NNZ]

    # Use only the upper triangle of the adjacency matrix
    areUpperTriangle <- row_indices < col_indices
    row_indices <- row_indices[areUpperTriangle]
    col_indices <- col_indices[areUpperTriangle]

    return(list(error_condition = error_condition,
                indices_zero_rows = zero_rows,
                row_indices = row_indices,
                col_indices = col_indices))

  }

  centers_coord <- basis_functions$centers_utm
  sepDistance <- basis_functions$sepDistance

  lb <- sepDistance - sepDistance * tol
  ub <- sepDistance + sepDistance * tol

  result <- CalcMatrixIndices(centers_coord, lb, ub, 1)

  if(result$error_condition == 1) {
    result <- CalcMatrixIndices(centers_coord, lb, ub, 10)
    if(result$error_condition == 1) {
      result <- CalcMatrixIndices(centers_coord, lb, ub, 100)
    }
    stop( sprintf("The number of non-zero elements in the adjacency matrix\n"),
          sprintf("is too large. There is probably an error in the basis functions.\n"),
          call. = FALSE )
  }

  if(result$error_condition == 2){
    return(result$indices_zero_rows)
  }


  # adjacency matrix in sparse form
  W <- Matrix::sparseMatrix(i = result$row_indices,
                            j = result$col_indices,
                            x = rep.int(1,length(result$row_indices)),
                            symmetric = TRUE)

  D_sparse <- Matrix::rowSums(W)

  X <- Matrix::Diagonal(x = 1.0 / sqrt(D_sparse))

  # Why is the vector of eigenvalues reversed (with function rev)?
  # So that vector lambda matches exactly the lambda that is calculated by the
  # functions written by M. Joseph.
  lambda <- rev(eigen(X %*% W %*% X, TRUE, only.values = TRUE)$values)

  return(list(nNeighborPairs = length(result$row_indices),
              W_sparse = cbind(result$row_indices, result$col_indices),
              D_sparse = D_sparse,
              lambda = lambda))
}

