#' @title Write a point shapefile
#'
#' @description
#' Write a point shapefile containing the predictions from the statistical modeling.
#'
#' @param object
#' Object of class Prediction.
#'
#' @param directory
#' Name of the directory that will contain the shapefile
#'
#' @param shapefile_name
#' Name of the shapefile to be written.
#'
#' @export
#'
write_point_shapefile <- function(object, directory, shapefile_name){

  spdf <- sp::SpatialPointsDataFrame(object$predictedValues[, c("long", "lat"), drop = FALSE],
                                     object$predictedValues[, c("easting", "northing", "Y_mean", "Y_sd",
                                                                "prob", "quantile"),
                                                            drop = FALSE],
                                     proj4string = sp::CRS(object$CRS_arg_longlat))
  rgdal::writeOGR(spdf, directory, shapefile_name, driver="ESRI Shapefile")

}

#' @title Write the predictions to raster files
#'
#' @description
#' Write the prediction quantities to raster files, for which
#' the file format is GeoTiff.
#'
#' @param object
#' Object of class Prediction.
#'
#' @details
#' Four raster files are written to directory \code{Rasterfiles}. File
#' Y_mean.tif contains the map of the process mean, file Y_sd.tif
#' contains the map of the
#' process standard deviation, file prob.tif contains the map of exceedance
#' probability, and file \code{quantile.tif} contains the map of the quantile.
#'
#' The projection of the rasters is Universal Transverse Mercator (utm).
#'
#' @export
#'
write_rasterfile <- function(object){

  if(!dir.exists("Rasterfiles")){
    dir.create("Rasterfiles")
  }

  for(quantity in c("Y_mean", "Y_sd", "prob", "quantile")){
    oRaster <- raster::rasterFromXYZ(object$predictedValues[, c("easting", "northing", quantity)],
                                     crs = object$CRS_arg_utm)

    filename <- paste0("Rasterfiles\\", quantity)
    raster::writeRaster(oRaster, filename, format = "GTiff", datatype = 'FLT8S')
  }

}

#' @title Read a polygon in a shapefile
#'
#' @description
#' The vertices of the polygon are geographic coordinates (that is,
#' latitudes and longitudes). For R-package BMNUS this polygon defines either
#' the domain or the region of interest.
#'
#' @param directory
#' Name of the directory containing the shapefiles
#'
#' @param shapefile_name
#' Name of the shapefile to be read.
#'
#' @return A list with the following components is returned.
#' @return \item{boundary}{R-language tibble storing the vertices of the
#' boundary, which are specified by longitude and latitude. The column
#' names are \code{long} and \code{lat}.}
#' @return \item{projection}{Character string storing the projection
#' of the coordinates in \code{boundary}. This projection format is
#' \code{proj4}, which is described in the
#' documentation of function sp::CRS.}
#'
#' @export
#'
read_polygon_shapefile <- function(directory, shapefile_name){

  # spatial polygon data frame
  spdf <- rgdal::readOGR(directory, shapefile_name)

  boundary <- tibble::as_tibble(spdf@polygons[[1]]@Polygons[[1]]@coords, .name_repair = "unique")
  colnames(boundary) <- c("long", "lat")

  return(list(boundary = boundary,
              projection = sp::proj4string(spdf)))
}

#' @title Read a point shapefile
#'
#' @description
#' Read a point shapefile of data for statistical modeling.
#'
#' @param directory
#' Name of the directory containing the shapefiles
#'
#' @param shapefile_name
#' Name of the shapefile to be read.
#'
#' @param isConcentration
#' Logical indicator specifying whether the measurements are concentrations.
#'
#' @param constSumValue
#' If the measurements are concentrations (argument \code{isConcentration}),
#' this argument is a real, positive-value scalar specifying the constant sum
#' value for those concentrations,
#'
#' @details
#' The shapefile must include at least the following:
#' \itemize{
#'   \item Measurements of the physical property that are to be modeled. The
#'   attribute name is \code{Value}.
#'   \item Indicators specifying whether measurements are censored. The
#'   attribute name is \code{IndicatorValue}.
#' }
#' An indictor is a character string: "left", "right", or "no".
#' String "left" means that the associated measurement is left censored;
#' string "right" means that the associated measurement is right censored;
#' string "no" means that the associated measurement is not censored.
#' If a measurement is left censored,
#' then the associated measurement value is the left censoring threshold.
#' Likewise, if a measurement is right censored,
#' the the associated measurement value is the right censoring threshold.
#' The data may have multiple, different censoring thresholds.
#'
#' If the measurements are concentrations, then they undergo the isometric
#' log-ratio transformation (Pawlowsky-Glahn and others, 2015). This
#' transformation requires a parameter called the constant sum value, which
#' is specified in argument \code{constSumValue}.
#' This argument is most easily explained with examples.
#' If the concentrations are reported in percent,
#' then argument \code{constSumValue} is 100; if they are reported in
#' milligrams per kilogram (or equivalently parts per million), then
#' argument \code{constSumValue} is 1000000; and if they are reported in
#' parts per billion, then argument \code{constSumValue} is 1000000000.

#' @return A R-language SpatialPointsDataFrame that contains the spatial
#' data. This container is defined in the sp package. The
#' SpatialPointsDataFrame contains the measurements in column \code{Value}
#' and the
#' censor indicators in column \code{IndicatorValue}. If the measurements
#' are concentrations, then the data in colum \code{Value} are the
#' isometric log-ratio transform of the concentrations. Also, the original
#' measurements (concentrations) are in column \code{OriginalValue}.
#'
#' @references
#' Pawlowsky-Glahn, V., Egozcue, J.J., and Tolosana-Delgado, R., 2015, Modeling
#' and analysis of compositional data: John Wiley and Sons, Ltd.
#'
#' @export
#'
read_point_shapefile <- function(directory, shapefile_name,
                                isConcentration = FALSE,
                                constSumValue = 1e6){

  # spatial points data frame
  spdf <- rgdal::readOGR(directory, shapefile_name)

  if(isConcentration){
    spdf@data$OriginalValue <- spdf@data$Value
    spdf@data$Value <- scaledLogit(spdf@data$Value, constSumValue = constSumValue)
  }

  return(spdf)
}
