#' @title Plot prediction locations
#'
#' @description
#' The plot of prediction locations (namely, the locations where the random
#' field is predicted) includes the domain boundary. This plot
#' is an incomplete map. To make the map useful, additional information
#' such as geographic boundaries should be added, and the map should be
#' projected properly. These tasks are accomplished using the functions in
#' library MapUtilities.
#'
#' @param object
#' Object of class Prediction.
#'
#' @param ...
#' Additional arguments to function MappingUtilities::Add_Points.
#'
#' @return A ggplot2 object to implement the previously-described plotting.
#'
#' @export
#'
plotPredictionLocations <- function(object, ...){

  ggplot2::ggplot() +
    # MappingUtilities::Add_Points(object$predictedValues[,c("long", "lat")],
    #                              ...) +
    MappingUtilities::Add_Points(object$predictedValues, ...) +
    MappingUtilities::Add_Path(object$domain_boundary, "blue")
}

#' @title Plot predicted quantites
#'
#' @description
#' This plot of the predicted quantities (namely, the values predicted at the
#' prediction locations) is an incomplete map. To make the map useful,
#' additional information
#' such as geographic information should be added, and the map should be
#' projected properly. These tasks are accomplished using the functions in
#' library MapUtilities.
#'
#' @param x
#' Object of class Prediction.
#'
#' @param plotPalette
#' Character vector containing the colors that are used to plot the predicted
#' quantities. This character vector is described in more detail in, for
#' example, the documentation for colorRampPalette.
#'
#' @param quantity
#' Character string containing the name of the predicted quantity that is
#' plotted. The choices are "Y_mean", "Y_sd", "prob", and "quantile".
#'
#' @param isPaletteScaled
#' Logical indicator specifying whether the colors are scaled by the quantiles
#' of the quantity that is plotted.
#'
#' @param ...
#' Additional arguments to function MappingUtilities::Add_Points.
#'
#' @details
#' Regarding argument \code{quantity}, "Y_mean" refers to
#' the process mean, "Y_sd" refers to the process standard deviation,
#' "prob" refers to the probability of exceeding \code{threshold}
#' (which is an argument to function Prediction), and "quantile"
#' corresponds to \code{probability} (which is another argument to function
#' Prediction).
#'
#' @return A ggplot2 object to implement the previously-described plotting.
#'
#' @export
#'
plot.Prediction <- function(x, plotPalette, quantity = "Y_mean",
                            isPaletteScaled = TRUE, ...){

  if(isPaletteScaled){
    tmp <- eval(parse(text = paste0("x$predictedValues$", quantity)))
    theQuantiles <-
      stats::quantile(tmp,
                      probs = seq(from = 0, to = 1,
                                  length.out = length(plotPalette)),
                      names = FALSE)
    values <- scales::rescale(theQuantiles)
  } else {
    values <- NULL
  }

  ggplot2::ggplot() +
    MappingUtilities::Add_Marked_Points(x$predictedValues, quantity,
                                        quantity, values, colors = plotPalette)

}

# Calc_pred_pts
#
# Calculation of the locations of the prediction points
#
# domain_boundary: Dataframe containing the longitudes and latitudes of the
#              points that define the boundary of the domain. The first column
#              of the dataframe lists the longitudes, and the second column
#              the latitudes. The column names are "long" and "lat".
# CRS_arg_longlat: arguments to function sp::CRS specfying the projection
#              for the argument conc_data
# CRS_arg_utm: arguments to function sp::CRS specifying the projection
#              for the utm coordinates
# n_pred_pts: Number of prediction points.
# seed: Random number seed.
#
# The seed may be set to a natural number to make the calculations reproducible.
#
# The locations of the prediction points are stored in a matrix. The first column lists
# the longitude, and the second lists the latitude. The datum for these
# coordinates is specified in the function arguments. Each row in the matrix
# pertains one prediction point.
#
# The function value is a list with two elements. The first is matrix
# pred_pts_longlat containing the longitude and latitude of the prediction
# points. The second is matrix pred_points_utm containing the easting and
# northing of the prediction points. The number of predictions points will be
# approximately equal to argument n_pred_pts; it will rarely equal n_pred_pts.
#
Calc_pred_points <- function(domain_boundary,
                             CRS_arg_longlat,
                             CRS_arg_utm,
                             n_pred_pts,
                             seed) {

  # recast the domain boundary as a spatial polygon container
  tmp <- as.matrix(domain_boundary)
  sp_domain_boundary <- sp::SpatialPolygons(list(sp::Polygons(list(sp::Polygon(tmp)), "1")),
                                        proj4string = sp::CRS(CRS_arg_longlat))


  # transform the lats/longs to utm
  sp_domain_boundary_utm <- sp::spTransform(sp_domain_boundary, sp::CRS(CRS_arg_utm) )


  # Calculate the prediction points in utm and latlong
  set.seed(seed)
  pred_pts_utm <- sp::spsample(sp_domain_boundary_utm, n_pred_pts, type = "regular")

  pred_pts_longlat <- sp::spTransform( pred_pts_utm, sp::CRS(CRS_arg_longlat))

  # only the coordinates of the prediction points are needed
  pred_pts_utm <- sp::coordinates(pred_pts_utm)
  colnames(pred_pts_utm) <- c("Easting", "Northing")

  pred_pts_longlat <- sp::coordinates(pred_pts_longlat)
  colnames(pred_pts_longlat) <- c("long", "lat")

  return(list(pred_pts_longlat = pred_pts_longlat,
              pred_pts_utm = pred_pts_utm))
}

# Predict
#
# Predict the process mean, the process standard deviation, the
# exceedance probability, and the quantile at the prediction points.
#
# pred_points: List returned by function Calc_pred_points. The list contains
#     the locations of the prediction points.
# basis_functions: List returned by function Calc_basis_func, which has the
#     complete documetantion of this list. Function CalcAdjacencyMatrix
#     uses two components of this list, namely centers_utm (which contains
#     the UTM coordinates of the centers of the basis functions) and
#     sepDistance (which is the distance, in meters, between adjacent centers).
# X: the measured values at the field sample locations.
# rawSamples: S4 class that is returned by rstan function sampling. It contains,
#     among other things, the samples of the posterior probability density
#     function.
# threshold: Quantile used to calculate an exceedance probability
# probability: Probability (e.g., 0.95) used to calculate the associated quantile.
#
# The exceedance probability is the probability of exceeding the threshold.
#
# The function value is a tibble containing, for each prediction location,
# the longitude, the latitude, the easting, the northing, the process mean,
# the process standard deviation, the probability of exceeding argument
# threshold, and the quantile associated with argument probability.
#
#
PredictValues <- function(pred_points, basis_functions, X, rawSamples,
                          threshold, probability, estNNZ = 1e7){

  postPdfSamples <- rstan::extract(rawSamples)

  basis_funcs_pred <- BasisFuncMatrix(basis_functions$centers_utm,
                                      pred_points$pred_pts_utm,
                                      basis_functions$sepDistance,
                                      estNNZ)

  tmp1 <- Matrix::tcrossprod(postPdfSamples$phi1, basis_funcs_pred)
  tmp2 <- exp(Matrix::tcrossprod(postPdfSamples$phi2, basis_funcs_pred))
  tmp3 <- Matrix::Diagonal(x = postPdfSamples$rho) %*% tmp2

  df <- tibble::tibble(
    # location of the prediction points
    long = pred_points$pred_pts_longlat[, "long"],
    lat = pred_points$pred_pts_longlat[, "lat"],
    easting = pred_points$pred_pts_utm[, "Easting"],
    northing = pred_points$pred_pts_utm[, "Northing"],

    # spatially-varying mean for the process model at the prediction points
    Y_mean = mean(X) + Matrix::colMeans(tmp1),

    # spatially-varying standard deviation for the process model at the
    # prediction points
    Y_sd = Matrix::colMeans(tmp3),

    # probability of exceeding the threshold
    prob = stats::pnorm(threshold, mean = Y_mean, sd = Y_sd, lower.tail = FALSE),

    # quantile
    quantile = stats::qnorm(probability, mean = Y_mean, sd = Y_sd, lower.tail = TRUE)
  )
  return(df)
}


#' @title Constructor of class to predict values throughout the domain.
#'
#' @description This constructor calculates the locations in the domain
#' at which various values
#' are predicted and then predicts the values at those locations. The predicted
#' values
#' are the process mean, the process standard deviation,
#' the probability of exceeding a specified threshold, and
#' the quantile associated with a specifed probability.
#'
#' @param domain_boundary
#' R-language dataframe containing the longitudes and latitudes of the
#'              points that define the boundary of the domain. The first column
#'              of the dataframe lists the longitudes, and the second column
#'              the latitudes. The column names are "long" and "lat".
#'
#' @param spatialData
#' Container with the spatial data and associated information.
#' See the User's Guide for a detailed description of this container.
#'
#' @param CRS_arg_utm
#' Character string containing the arguments to function sp::CRS that
#' specify the projection, in universal transverse mercator (utm) coordinates,
#' for arguments domain_boundary.
#'
#' @param basis_functions
#' R-language list containing the basis functions at the selected spacing
#' between function centers. This list is completely described in the
#' documentation for function BasisFunctions.
#'
#' @param rawSamples
#' S4 class returned by function rstan::sampling, which contains samples
#' of the posterior probability density function.
#'
#' @param threshold
#' Real-valued scalar. The probability of exceeding this threshold is calculated
#' at all prediction locations. If the original data are concentrations, then this
#' threshold must be transformed with the isometric log-ratio transform before
#' it is passed to this function.
#'
#' @param probability
#' Real-valued scalar between 0 and 1. The quantile at this probability is
#' calculated at all prediction locations.
#'
#' @param n_pred_pts
#' Number of prediction locations
#'
#' @param seed
#' Random number seed. Calculating the locations at which the values
#' are predicted is partly random, so setting this seed makes the calculations
#' reproducible.
#'
#' @param directory_name
#' Character string containing the name of the directory that will store
#' the files with the predictions.
#'
#' @details
#' The prediction locations are stored in file \code{Pred_points.dat} in
#' directory \code{directory_name}. The predicted values likewise are stored in
#' file \code{PredictedValues.dat} in directory \code{directory_name}.
#'
#' @return A list with the following components is returned.
#' @return \item{domain_boundary}{Argument \code{domain_boundary}.}
#' @return \item{CRS_arg_longlat}{Argument \code{CRS_arg_longlat}.}
#' @return \item{CRS_arg_utm}{Argument \code{CRS_arg_utm}.}
#' @return \item{threshold}{Argument \code{threshold}.}
#' @return \item{probability}{Argument \code{probability}.}
#' @return \item{n_pred_pts}{Argument \code{n_pred_pts}.}
#' @return \item{seed}{Argument \code{seed}.}
#' @return \item{directory_name}{Argument \code{directory_name}.}
#' @return \item{predictedValues}{R-language tibble containing, at each
#' prediction location,
#' \itemize{
#'   \item \code{long}, the longitude
#'   \item \code{lat}, the latitude
#'   \item \code{easting}, the utm easting
#'   \item \code{northing}, the utm northing
#'   \item \code{Y_mean}, the process mean
#'   \item \code{Y_sd}, the process standard deviation
#'   \item \code{prob}, the probability of exceeding argument \code{threshold}
#'   \item \code{quantile}, the quantile associated with
#'          argument \code{probability}
#' }
#' The datum for the longitudes and latitudes
#' is specifed in argument \code{CRS_arg_longlat}. The datum for the easthings
#' and northings is is specifed in argument \code{CRS_arg_utm}.}
#' @return \item{call}{Function call.}
#'
#' @export
#'
Prediction <- function(domain_boundary,
                       spatialData,
                       CRS_arg_utm,
                       basis_functions,
                       rawSamples,
                       threshold,
                       probability = 0.95,
                       n_pred_pts = 10000L,
                       seed = 777,
                       directory_name = "Predictions"){

  if(!dir.exists(directory_name)){
    dir.create(directory_name)
  }

  CRS_arg_longlat <- sp::proj4string(spatialData)

  pred_points <- Calc_pred_points(domain_boundary,
                                  CRS_arg_longlat,
                                  CRS_arg_utm,
                                  n_pred_pts, seed)

  predictedValues <- PredictValues(pred_points, basis_functions,
                                   spatialData@data$Value,
                                   rawSamples, threshold, probability)

  save(pred_points, file = paste0(directory_name, "\\Pred_points.dat"))
  save(predictedValues, file = paste0(directory_name, "\\PredictedValues.dat"))

  rval <- list(domain_boundary = domain_boundary,
               CRS_arg_longlat = CRS_arg_longlat,
               CRS_arg_utm = CRS_arg_utm,
               threshold = threshold,
               probability = probability,
               n_pred_pts = n_pred_pts,
               seed = seed,
               directory_name = directory_name,
               predictedValues = predictedValues,
               call = sys.call())

  class(rval) <- "Prediction"

  return(rval)


}
