#' @title Plot the results of the simple cross validation test
#'
#' @description There are two plots: One plot is the deviance versus distance
#' between basis function centers, and the other plot is mean squared error
#' versus distance between basis function centers.
#'
#' @param x
#' Object of class SimpleCrossValidation
#'
#' @param jitter_width
#' Real-valued scalar specifying the spread of the random noise. See Details.
#'
#' @param ...
#' Required for compatibility with generic definition
#'
#' @export
#'
plot.SimpleCrossValidation <- function(x, jitter_width = 0.005, ...) {

  p1 <- ggplot2::ggplot(data = x$results, mapping = ggplot2::aes(x = spacing, y = deviance)) +
    ggplot2::geom_point(shape = 16) +
    ggplot2::scale_x_continuous(name = "Distance between basis function centers, in kilometers",
                                trans = "log10",
                                breaks = sort(x$results$spacing)) +
    ggplot2::scale_y_continuous(name = "Deviance")

  p2 <- ggplot2::ggplot(data = x$results, mapping = ggplot2::aes(x = spacing, y = mse)) +
    ggplot2::geom_point(shape = 16) +
    ggplot2::scale_x_continuous(name = "Distance between basis function centers, in kilometers",
                                trans = "log10",
                                breaks = sort(x$results$spacing)) +
    ggplot2::scale_y_continuous(name = "Mean squared error")

  grid::grid.newpage()
  grid::pushViewport(grid::viewport(layout=grid::grid.layout(2,1)))
  plot(p1, vp=grid::viewport(layout.pos.row=1, layout.pos.col=1))
  plot(p2, vp=grid::viewport(layout.pos.row=2, layout.pos.col=1))
}

#' @title Summarize the results of the simpe cross validation test
#'
#' @description Print a table that summarizes, at each of the spacings,
#' the two measures of predictive accuracy: the deviance, and
#' the mean squared error.
#'
#' @param object
#' Object of class SimpleCrossValidation
#'
#' @param ...
#' Required for compatibility with generic definition
#'
#' @export
#'
summary.SimpleCrossValidation <- function(object, ...) {
  cat( sprintf( "\n#########################################################################\n"))
  print(object$results)
  cat( sprintf( "#########################################################################\n"))
}


# CalcPredictiveAccuracy_SXV
#
# Calculate the predictive accuracy for a simple cross validation test at
# one spacing between basis function centers.
#
# filename: Name of the file with the basis functions.
#
# training_spatialData:
# Container with the training measurements, which are a type of spatial data.
# These data are used to training the statistical model, namely estimate the
# parameters in the model.
# See the User's Guide for a detailed description of this container.
#
# validation_spatialData:
# Container with the validation measurements, which also are a type of spatial
# data. These
# data are used to test how well the statistical model can predict values
# of the physical property that is being modeled.
# See the User's Guide for a detailed description of this container.
#
# me_sd: Standard deviation of the measurement error.
#
# BMNUS_sm: The compiled stan program BMNUS.
#
# estNNZ: Estimated number of non-zero elements in the matrix of
# basis fuction values.
#
# Value
# A named vector containing the deviance and the mean squared error, which
# are calculated for the validation measurements.
#
# Reference:
# Gelman, A., Carlin, J.B., Stern, H.S., Dunson, D.B., Vehtari, A., and
# Rubin, D.B., 2014, Bayesian data analysis (3d ed.): Boca Raton, Fla.,
# CRC Press, section 7.1 (p. 166-178).
CalcPredictiveAccuracy_SXV <- function(filename,
                                       training_spatialData,
                                       validation_spatialData_utm,
                                       BMNUS_sm,
                                       estNNZ){

  require(Matrix, quietly = TRUE)
  require(rstan, quietly = TRUE)

  # loads variable basis_functions, which is an R-language list
  load(filename)

  # Transform matrix basisFuncs from "dictionary-of-keys" format  to
  # "compressed-row-storage" format.
  csrBasisFuncs <- as(basis_functions$basisFuncs,
                      "RsparseMatrix")

  # Calculate quantities that are needed for the sparse CAR model;
  # these quantities are derived from the adjacency matrix
  carQuantities <- SparseCarQuantities(basis_functions)

  stanData <- list( N = nrow(training_spatialData),
                    M = ncol(csrBasisFuncs),
                    UU = csrBasisFuncs@p + 1,
                    VV = csrBasisFuncs@j + 1,
                    WW = csrBasisFuncs@x,
                    nNeighborPairs = carQuantities$nNeighborPairs,
                    W_sparse = carQuantities$W_sparse,
                    D_sparse = carQuantities$D_sparse,
                    lambda = carQuantities$lambda,
                    X = training_spatialData@data$Value,
                    me_sd = training_spatialData@data$me_sd,
                    areLeftCensored = as.integer(training_spatialData@data$IndValue == "left"),
                    betaPar = c(2.5, 1.2),
                    gammaPar = c(2, 0.3),
                    cauchyPar = 3)

  # Function gen_int initializes some variables for the sampling
  gen_init <- function(){
    phi1 <- stats::rnorm(stanData$M, mean = 0.0, sd = 1e-6)
    phi2 <- stats::rnorm(stanData$M, mean = 0.0, sd = 1e-6)
    return(list(phi1 = phi1, phi2 = phi2))
  }

  # Sample the posterior pdf
  rawSamples <- rstan::sampling(BMNUS_sm, data = stanData, chains = 1,
                         iter = 2500, warmup = 500,
                         refresh = 500,
                         init = gen_init,
                         pars=c("rho", "phi1", "phi2"))

  # compute three measures of predictive accuracy.
  # B_test <- basis_functions$basisFuncs
  centers_utm <- sp::SpatialPoints(basis_functions$centers_utm,
                                   proj4string = sp::CRS(basis_functions$CRS_arg_utm))
  sample_locations_utm <- sp::SpatialPoints(sp::coordinates(validation_spatialData_utm),
                                            proj4string = sp::CRS(sp::proj4string(validation_spatialData_utm)))
  B_test <- BasisFuncMatrix(centers_utm,
                            sample_locations_utm,
                            basis_functions$sepDistance, estNNZ)
  X_test <- validation_spatialData_utm@data$Value
  areLeftCensored_test <- as.integer(validation_spatialData_utm@data$IndValue == "left")
  me_sd_test <- validation_spatialData_utm@data$me_sd

  # Compute Y_mean and Y_sd for each realization
  # of random variables phi1, phi2, and rho. Then average Y_mean and Y_sd
  # to obtain final values for Y_mean and Y_sd.
  postPdfSamples <- rstan::extract(rawSamples)

  tmp1 <- Matrix::tcrossprod(postPdfSamples$phi1, B_test)
  Y_mean = mean(stanData$X) + Matrix::colMeans(tmp1)

  tmp2 <- exp(Matrix::tcrossprod(postPdfSamples$phi2, B_test))
  tmp3 <- Matrix::Diagonal(x = postPdfSamples$rho) %*% tmp2
  Y_sd = Matrix::colMeans(tmp3)

  logLikelihood <- 0.0
  squared_error <- numeric(length(X_test))
  weighted_squared_error <- numeric(length(X_test))

  for(j in 1:length(X_test)){
    sigma <- sqrt(Y_sd[j]^2 + (me_sd_test[j])^2)
    if(areLeftCensored_test[j] == 1){
      logLikelihood <- logLikelihood +
        stats::pnorm(X_test[j], mean = Y_mean[j], sd = sigma, log = TRUE)
    } else {
      logLikelihood <- logLikelihood +
        stats::dnorm(X_test[j], mean = Y_mean[j], sd = sigma, log = TRUE)
    }
    squared_error[j] <- (X_test[j] - Y_mean[j])^2
    weighted_squared_error[j] <- (X_test[j] - Y_mean[j])^2 / sigma^2
  }

  return(c(deviance = -2.0 * logLikelihood,
                mse = mean(squared_error)))
}



#' @title Constructor of class to perform simple cross validation test
#'
#' @description
#' This constructor performs a simple cross validation test for one or more
#' spacings
#' between basis function centers. The qualifier "simple" means that
#' validation data is available and is used for the cross validation.
#' The test is performed using parallel processing to reduce computation time.
#'
#' @param training_spatialData
#' Container with the training measurements, which are a type of spatial data.
#' These data are used to training the statistical model, namely estimate the
#' parameters in the model.
#' See the User's Guide for a detailed description of this container.
#'
#' @param validation_spatialData
#' Container with the validation measurements, which also are a type of spatial
#' data. These
#' data are used to test how well the statistical model can predict values
#' of the physical property that is being modeled.
#' See the User's Guide for a detailed description of this container.
#'
#' @param BMNUS_sm
#' The compiled stan program BMNUS.
#'
#' @param oBasisFunctions
#' Object of class BasisFunctions.
#'
#' @param nCpuCores
#' Number of processor (cpu) cores that are used for the parallel calculations.
#' This parameter should be less than the number of logical cores.
#'
#' @details
#' The duration of the computations depends upon many factors, including
#' the number of basis functions, the number of training measurements,
#' the number of validation measurements,
#' and the speed of the computer. Thus, predicting the
#' duration is practically impossible.
#'
#' @return A list with the following components is returned.
#' @return \item{nCpuCores}{Argument \code{nCpuCores}.}
#' @return \item{results}{R-language tibble, for which the number of rows
#' equals the  number of spacings between basis function centers. The
#' first column lists the spacings, in kilometers, between basis function
#' centers. The second column lists the deviances, and the third column lists
#' the mean squared errors. The deviance and mean squared error
#' are measures of predictive accuracy in the cross validation test
#' (Gelman and others, 2014, p. 166-178).}
#' @return \item{call}{Function call.}
#'
#' @references
#' Gelman, A., Carlin, J.B., Stern, H.S., Dunson, D.B., Vehtari, A., and
#' Rubin, D.B., 2014, Bayesian data analysis (3d ed.): Boca Raton, Fla.,
#' CRC Press.
#'
#' @export
#'
SimpleCrossValidation <- function(training_spatialData,
                                  validation_spatialData,
                                  BMNUS_sm,
                                  oBasisFunctions,
                                  nCpuCores){

  load(oBasisFunctions$file_info$filenames[1])
  CRS_arg_utm <- basis_functions$CRS_arg_utm
  validation_spatialData_utm <- sp::spTransform(validation_spatialData,
                                                sp::CRS(CRS_arg_utm))

  CL <- parallel::makeCluster(nCpuCores)

  results <- parallel::parSapplyLB(CL,
                                 oBasisFunctions$file_info$filenames,
                                 CalcPredictiveAccuracy_SXV,
                                 training_spatialData,
                                 validation_spatialData_utm,
                                 BMNUS_sm,
                                 oBasisFunctions$estNNZ)

  parallel::stopCluster(CL)

  temp <- tibble::tibble(spacing = oBasisFunctions$file_info$center_spacings,
                         deviance = results["deviance", ],
                         mse = results["mse", ])

  rval <- list(nCpuCores = nCpuCores,
               results = temp,
               call = sys.call())

  class(rval) <- "SimpleCrossValidation"

  return(rval)
}

