#' @title Plot the spatial data along the transect
#'
#' @description
#' The values of the field samples that are included in the transect
#' are plotted as a function of distance along the transect.
#'
#' @param x
#' Object of class ViewTransect.
#'
#' @param yAxisLabel
#' Label for the y axis
#'
#' @param isCurveAdded
#' Logical variable indicating whether a smooth curve is added to the plot.
#'
#' @param ...
#' Additional arguments to functions that calculate and plot the smooth curve,
#' namely functions loess and ggplot2::geom_smooth.
#'
#' @export
#'
plot.ViewTransect <- function(x, yAxisLabel,
                              isCurveAdded = TRUE, ...){

  areInTransect <- x$areInTransect

  df <- tibble::tibble(distance = x$distanceAlongTransect,
                       X = x$Value)

  p1 <- ggplot2::ggplot(data = df,
                        mapping = ggplot2::aes(x = distance, y = X)) +
    ggplot2::geom_point(color = "black")

  if(isCurveAdded){
    p1 <- p1 +
      ggplot2::geom_smooth(..., color = "red", size = 1)
  }
  p1 <- p1 +
    ggplot2::scale_x_continuous(name = "Distance, in kilometers") +
    ggplot2::scale_y_continuous(name = yAxisLabel) +
    ggplot2::theme_bw() +
    ggplot2::theme(panel.grid.major = ggplot2::element_blank(),
                   panel.grid.minor = ggplot2::element_blank()) +
    ggplot2::theme(plot.title=ggplot2::element_text(hjust = 0,face = "bold.italic", size = 10))

  print(p1)

}

#' @title Constructor of class to view a transect
#'
#' @description
#' This constructor performs various calculations that are needed to view a
#' transect, which is used to determine a
#' suitable spacing for the basis function centers.
#'
#' @param spatialData
#' Container with the spatial data and associated information.
#' See the User's Guide for a detailed description of this container.
#'
#' @param maxOffset
#' Real-valued scalar that specifies the maximum offset---namely, the
#' maximum distance (in kilometers) between a field sample and the transect.
#' Field
#' samples for which the offset is less than maxOffset are included in the
#' transect; otherwise, they are not.
#'
#' @param end_points
#' R-language dataframe containing the locations of the two ends of the
#' transect. Each row pertains one end.
#' The locations are specified with longitude and latitude; the column headings
#' in the dataframe are "long" and "lat". The datum for the longitudes and
#' latitudes must be the same as that used for argument \code{spatialData}.
#'
#' @param CRS_arg_utm
#' Character string containing the arguments to function sp::CRS that
#' specify the projection, in universal transverse mercator (utm) coordinates,
#' for arguments \code{spatialData} and \code{end_points}.
#'
#' @return A list with the following components is returned.
#' @return \item{maxOffset}{Argument \code{maxOffset}.}
#' @return \item{end_points}{Argument \code{end_points}.}
#' @return \item{CRS_arg_longlat}{Character string containing the arguments to
#' function sp::CRS that specify the projection for longitudes and latitudes
#' for arguments \code{spatialData} and \code{end_points.}}
#' @return \item{CRS_arg_utm}{Argument \code{CRS_arg_utm}.}
#' @return \item{areInTransect}{Logical vector indicating whether the field
#' samples in \code{spatialData} are included in the transect.}
#' @return \item{distanceAlongTransect}{Real-valued vector containing the
#' distances along the transect of just the field samples that are in the
#' transect.}
#' @return \item{distanceFromTransect}{Real-valued vector containing the
#' distances from the transect (namely, the offset perpendicular to the
#' transect) of just the field samples that are in the transect.}
#' @return \item{sampleLocations_longlat}{R-language tibble containing the
#' longitudes and latitudes of just the field samples that are included in
#' the transect.}
#' @return \item{Value}{Real-valued vector containing the measured values of
#' just the field samples that are in the transect.}
#' @return \item{call}{Function call.}
#'
#' @export
#'
ViewTransect <- function(spatialData, maxOffset, end_points,
                          CRS_arg_utm){

  rval <- CheckTransect(spatialData, maxOffset, end_points, CRS_arg_utm)
  rval$call <- sys.call()

  class(rval) <- "ViewTransect"

  return(rval)

}
