/*
Argument        Meaning
N               Number of data values.
M               Number of basis functions.

UU              Array of integer indices indicating where in vector WW a
                given row’s values start.
VV              Array of integer column indices for the values.
WW              Vector containing the non-zero elements of the matrix
                containing the basis functions.

nNeighborPairs  Integer-valued scalar. Number of neighbor pairs in the adjacency matrix
W_sparse        Integer-valued matrix. The first column lists the row
                indices of the non-zero elements in the upper triangle of the adjacency
                matrix. Only the upper triangle is used because the adjacency matrix is
                symmetric. The second column lists the associated column indices of the
                adjacency matrix.
D_sparse        Real-valued vector. This vector is the diagonal of
                diagonal matrix D that is used in sparse CAR formulation. See
                reference by Joseph.
lambda          Real-valued vector. This vector comprises the
                eigenvalues of matrix D^(-1/2) W D^(-1/2) that is used in sparse CAR
                formulation. See the reference by Joseph.
X               data
me_sd           standard deviation of measurement error
areLeftCensored Indicator. If X[i] is left-censored, then areLeftCensored[i] = 1,
                and X[i] is set to the left-censored threshold. If X[i] is not
                left-censored, then areLeftCensored[i] = 0.
betaPar         Two shape parameters characterizing the beta pdf.
                betaPar[1] and betaPar[2] are typically designated alpha and
                beta.
gammaPar        Two parameters characterizing the gamma pdf.
                gammaPar[1] contains shape parameter alpha, and
                gammaPar[2] contains inverse scale beta.
cauchyPar       Scale parameter for the Cauchy pdf.

Details

Arguments UU, VV, and WW store, in compressed row storage, the matrix
containing the basis functions. This storage format is described in the
chapter "Sparse Matrix Operations" in the  book "Stan Functions Reference".

The beta pdf is a prior pdf for parameter alpha in the conditional
autoregressive (CAR) model. This alpha controls the spatial dependence among
the basis functions. If alpha = 0, then there is no spatial dependence; if
alpha = 1, there is very strong spatial dependence. The values between these
two extremes are specified by the beta pdf. Because the two extremes should be
avoided, both values in betaPar should be greater than 1. Furthermore,
because some spatial dependence is desirable, betaPar[1] should be greater than
betaPar[2]. Suitable values for betaPar might be 2.5 and 1.2.

The gamma pdf is the prior pdf for parameter tau in the CAR model. This tau
represents the precision in the normal conditional pdf. There are two extremes
to be avoided: very small (near 0) and very large values of tau.
(A large value depends on the specific problem, so it cannot be precisely
defined.) These extremes can be avoided with the gamma pdf, if suitable values
of its parameters gammaPar[1] and gammaPar[2] are selected. The probability of a
value near 0 diminishes as gammaPar[1] increases above 1. Thus, gammaPar[1]
should always be greater than 1. The probability of a very large value
diminishes as gammaPar[2] increases above 0. Suitable values for gammaPar
might be 2 and 0.3.

The truncated Cauchy pdf is the prior pdf for the parameter rho. This rho
represents the average standard deviation throughout the domain. The pdf for rho
must statisfy two criteria. First, it must permit only positive values for rho.
Second, it must permit rho to have a wide range of values because usually little
is known about the average standard deviation before the statistical modeling.
To satisfy these two requirements, rho is chosen to have a truncated Cauchy
distribution. To understand this distribution, consider a Cauchy distribution
with a center of 0 and a scale parameter of cauchyPar. Then truncate the Cauchy
distribution at 0, removing the negative-valued part. The resulting distribution
ensures that the standard deviation is positive-valued, which satisfies the
first criterion. A suitable value for scale parameter cauchypar might be 3,
which satisfies the second criterion.

References
Joseph, Max, Exact sparse CAR models in Stan, Internet web page available at
http://mc-stan.org/documentation/case-studies/mbjoseph-CARStan.html,
last accessed March 2017.

Stan Development Team, Stan Functions Reference, version 2.19, available at
https://mc-stan.org

*/

functions {
  /**
  * Return the log probability of a proper conditional autoregressive (CAR) prior

  * with a sparse representation for the adjacency matrix
  *
  * param phi Vector containing the parameters with a CAR prior
  * param tau Precision parameter for the CAR prior (real)
  * param alpha Dependence (usually spatial) parameter for the CAR prior (real)
  * param W_sparse Sparse representation of adjacency matrix (int array)
  * param n Length of phi (int)
  * param W_n Number of adjacent pairs (int)
  * param D_sparse Number of neighbors for each location (vector)
  * param lambda Eigenvalues of D^{-1/2}*W*D^{-1/2} (vector)
  *
  * return Log probability density of CAR prior up to additive constant

  Developed by M. Joseph
  */
  real sparse_car_lpdf(vector phi, real tau, real alpha,
    int[,] W_sparse, vector D_sparse, vector lambda, int n, int W_n) {
      row_vector[n] phit_D; // phi' * D
      row_vector[n] phit_W; // phi' * W
      vector[n] ldet_terms;

      phit_D = (phi .* D_sparse)';
      phit_W = rep_row_vector(0, n);
      for (i in 1:W_n) {
        phit_W[W_sparse[i, 1]] = phit_W[W_sparse[i, 1]] + phi[W_sparse[i, 2]];
        phit_W[W_sparse[i, 2]] = phit_W[W_sparse[i, 2]] + phi[W_sparse[i, 1]];
      }

      for (i in 1:n) ldet_terms[i] = log1m(alpha * lambda[i]);
      return 0.5 * (n * log(tau)
                    + sum(ldet_terms)
                    - tau * (phit_D * phi - alpha * (phit_W * phi)));
  }

  // Developed by M. Joseph
  // With the current implementation of program BMNUS, W_sparse is pre-computed.
  // So, this function is not used.
  int[,] Calc_W_sparse(matrix W, int W_n, int M){
    int W_sparse[W_n, 2] ;
    int counter ;

    counter = 1;
    // loop over upper triangular part of W to identify neighbor pairs
    for (i in 1:(M - 1)) {
      for (j in (i + 1):M) {
        if (W[i, j] == 1.0) {
          // the original code was W[i, j] == 1
          W_sparse[counter, 1] = i;
          W_sparse[counter, 2] = j;
          counter = counter + 1;
        }
      }
    }
    return W_sparse ;
  }

  // Developed by M. Joseph
  // With the current implementation of program BMNUS, D_sparse is pre-computed.
  // So, this function is not used.
  vector Calc_D_sparse(matrix W, int M){
    vector[M] D_sparse ;

    for (m in 1:M) D_sparse[m] = sum(W[m]);

    return D_sparse ;
  }

  // Developed by M. Joseph
  // With the current implementation of program BMNUS, lambda is pre-computed.
  // So, this function is not used.
  vector Calc_lambda(matrix W, vector D_sparse, int M){
    vector[M] invsqrtD ;
    vector[M] lambda ;

    for (m in 1:M) {
      invsqrtD[m] = 1 / sqrt(D_sparse[m]);
    }

    lambda = eigenvalues_sym(quad_form(W, diag_matrix(invsqrtD))) ;

    return lambda ;
  }

}
data{
  int<lower=1> N ;
  int<lower=1> M ;
  int UU[N+1] ;
  int VV[UU[N+1] - 1] ;
  vector[UU[N+1] - 1] WW ;
  int<lower=1> nNeighborPairs ;
  int W_sparse[nNeighborPairs, 2] ;
  vector<lower=1>[M] D_sparse ;
  vector[M] lambda ;
  vector[N] X ;
  vector<lower=0>[N] me_sd ;
  vector<lower=0,upper=1>[N] areLeftCensored ;
  vector<lower=1>[2] betaPar ;
  vector<lower=0>[2] gammaPar ;
  real<lower=0> cauchyPar ;
}
transformed data{
  // mu             mean for the entire domain
  real mu = mean(X) ;
}
parameters{
  /*
  phi1             weights for the basis functions that are used for the mean
  phi2             weights for the basis functions that are used for the
                   standard deviation
  tau1 and alpha1  parameters in the CAR model for phi1
  tau2 and alpha2  parameters in the CAR model for phi2
  rho              standard deviation for the entire domain
  */
  vector[M] phi1 ;
  vector[M] phi2 ;
  real<lower=0> tau1 ;
  real<lower=0,upper=1> alpha1 ;
  real<lower=0> tau2 ;
  real<lower=0,upper=1> alpha2 ;
  real<lower=0> rho ;
}
transformed parameters{
  /*
  Y_mean          mean for the process model
  Y_sd            standard deviation for the process model
  */
  vector[N] Y_mean = mu + csr_matrix_times_vector(N, M, WW, VV, UU, phi1) ;
  vector[N] Y_sd = rho * exp(csr_matrix_times_vector(N, M, WW, VV, UU, phi2)) ;
}
model{
  // prior pdfs (parameter model)
  target += beta_lpdf(alpha1| betaPar[1], betaPar[2]) ;
  target += beta_lpdf(alpha2| betaPar[1], betaPar[2]) ;
  target += gamma_lpdf(tau1 | gammaPar[1], gammaPar[2]) ;
  target += gamma_lpdf(tau2 | gammaPar[1], gammaPar[2]) ;
  target += cauchy_lpdf(rho | 0, cauchyPar) ;

  // CAR model for the weights of the basis functions (parameters phi1 and phi2)
  target += sparse_car_lpdf(phi1 | tau1, alpha1, W_sparse, D_sparse,
            lambda, M, nNeighborPairs) ;
  target += sparse_car_lpdf(phi2 | tau2, alpha2, W_sparse, D_sparse,
            lambda, M, nNeighborPairs) ;

  // likelihood function (combination of the process and data models)
  for(n in 1:N){
    real sigma = sqrt(square(Y_sd[n]) + square(me_sd[n])) ;
    if(areLeftCensored[n] == 1){
      target += normal_lcdf(X[n] | Y_mean[n], sigma) ;
    } else {
      target += normal_lpdf(X[n] | Y_mean[n], sigma) ;
    }
  }
}
